(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2014, 2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2014, 2015)
License: LGPL (see file COPYING.LESSER)
*)

section \<open>Subsumption and Encompassment\<close>

theory Encompassment
imports 
  "../Rewriting/Renaming_Interpretations"
begin

no_notation
  less_multiset (infix "<#" 50)

subsection \<open>Subsumption (the ``instance of'' relation)\<close>

text \<open>Syntax for subsumption and literal similarity (also as sets).\<close>
consts
  SUBSUMESEQ :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<le>\<cdot>" 50)
  SUBSUMES :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "<\<cdot>" 50)
  LITSIM :: "'a \<Rightarrow> 'a \<Rightarrow> bool" (infix "\<doteq>" 50)

abbreviation (input) INSTANCEQ (infix "\<cdot>\<ge>" 50)
where
  "x \<cdot>\<ge> y \<equiv> y \<le>\<cdot> x"

abbreviation (input) INSTANCE (infix "\<cdot>>" 50)
where
  "x \<cdot>> y \<equiv> y <\<cdot> x"

abbreviation SUBSUMESEQ_SET ("{\<le>\<cdot>}")
where
  "{\<le>\<cdot>} \<equiv> {(x, y). x \<le>\<cdot> y}"

abbreviation SUBSUMES_SET ("{<\<cdot>}")
where
  "{<\<cdot>} \<equiv> {(x, y). x <\<cdot> y}"

abbreviation INSTANCEEQ_SET ("{\<cdot>\<ge>}")
where
  "{\<cdot>\<ge>} \<equiv> {(x, y). y \<le>\<cdot> x}"

abbreviation INSTANCE_SET ("{\<cdot>>}")
where
  "{\<cdot>>} \<equiv> {(x, y). y <\<cdot> x}"

locale subsumable =
  fixes subsumeseq :: "'a \<Rightarrow> 'a \<Rightarrow> bool"
  assumes refl: "subsumeseq x x"
    and trans: "subsumeseq x y \<Longrightarrow> subsumeseq y z \<Longrightarrow> subsumeseq x z"
begin

adhoc_overloading
  SUBSUMESEQ subsumeseq

definition "subsumes t s \<longleftrightarrow> t \<le>\<cdot> s \<and> \<not> s \<le>\<cdot> t"

definition "litsim s t \<longleftrightarrow> s \<le>\<cdot> t \<and> t \<le>\<cdot> s"

adhoc_overloading
  SUBSUMES subsumes and
  LITSIM litsim

lemma litsim_refl [simp]:
  "s \<doteq> s"
  by (auto simp: litsim_def refl)

lemma litsim_sym:
  "s \<doteq> t \<Longrightarrow> t \<doteq> s"
  by (auto simp: litsim_def)

lemma litsim_trans:
  "s \<doteq> t \<Longrightarrow> t \<doteq> u \<Longrightarrow> s \<doteq> u"
  by (auto simp: litsim_def dest: trans)

end

sublocale subsumable \<subseteq> subsumption!: preorder "op \<le>\<cdot>" "op <\<cdot>"
  by (unfold_locales) (auto simp: subsumes_def refl elim: trans)

inductive subsumeseq_term :: "('a, 'b) term \<Rightarrow> ('a, 'b) term \<Rightarrow> bool"
where
  [intro]: "t = s \<cdot> \<sigma> \<Longrightarrow> subsumeseq_term s t"

adhoc_overloading
  SUBSUMESEQ subsumeseq_term

lemma subsumeseq_termE [elim]:
  assumes "s \<le>\<cdot> t"
  obtains \<sigma> where "t = s \<cdot> \<sigma>"
  using assms by (cases)

lemma subsumeseq_term_refl:
  fixes t :: "('a, 'b) term"
  shows "t \<le>\<cdot> t"
  by (rule subsumeseq_term.intros [of t t Var]) simp

lemma subsumeseq_term_trans:
  fixes s t u :: "('a, 'b) term"
  assumes "s \<le>\<cdot> t" and "t \<le>\<cdot> u"
  shows "s \<le>\<cdot> u"
proof -
  obtain \<sigma> \<tau>
    where [simp]: "t = s \<cdot> \<sigma>" "u = t \<cdot> \<tau>" using assms by fastforce
  show ?thesis
    by (rule subsumeseq_term.intros [of _ _ "\<sigma> \<circ>\<^sub>s \<tau>"]) simp
qed

interpretation term_subsumable: subsumable subsumeseq_term
  by (default) (force simp: subsumeseq_term_refl dest: subsumeseq_term_trans)+

adhoc_overloading
  SUBSUMES term_subsumable.subsumes and
  LITSIM term_subsumable.litsim

lemma subsumeseq_term_iff:
  "s \<cdot>\<ge> t \<longleftrightarrow> (\<exists>\<sigma>. s = t \<cdot> \<sigma>)"
  by auto

lemma listsim_equiv_on_terms:
  "equiv T {(x, y). x \<doteq> y \<and> x \<in> T \<and> y \<in> T}"
  by (rule equivI)
     (auto simp: refl_on_def sym_def trans_def
           dest: term_subsumable.litsim_sym term_subsumable.litsim_trans)

lemma equivp_litsim:
  "equivp (op \<doteq> :: ('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool)"
  by (rule equivpI) (auto simp: reflp_def symp_def transp_def
                          dest: term_subsumable.litsim_sym term_subsumable.litsim_trans)

lemma litsim_term_iff:
  fixes s t :: "('f, 'v :: infinite) term"
  shows "s \<doteq> t \<longleftrightarrow> (\<exists>p. p \<bullet> s = t)"
proof
  assume "\<exists>p. p \<bullet> s = t"
  then obtain p where "p \<bullet> s = t" ..
  then have "t = s \<cdot> sop p" and "s = t \<cdot> sop (-p)" by auto
  from subsumeseq_term.intros [OF this(1)] and subsumeseq_term.intros [OF this(2)]
    show "s \<doteq> t" by (auto simp: term_subsumable.litsim_def)
next
  assume "s \<doteq> t"
  then obtain \<sigma> \<tau>
    where "s \<cdot> \<sigma> = t" and "t \<cdot> \<tau> = s" by (fastforce simp: term_subsumable.litsim_def)
  from variants_imp_renaming [OF this] obtain f
    where "bij f" and "finite {x. f x \<noteq> x}" and [simp]: "t = s \<cdot> (Var \<circ> f)" by blast
  then have "f \<in> perms" by (auto simp: perms_def)
  then have "Abs_perm f \<bullet> s = t" by (auto simp: term_apply_subst_Var_Abs_perm)
  then show "\<exists>p. p \<bullet> s = t" ..
qed

definition subsumeseq_trs :: "('f, 'v :: infinite) trs \<Rightarrow> ('f, 'v) trs \<Rightarrow> bool"
where
  "subsumeseq_trs R S \<longleftrightarrow> (\<forall>(l, r) \<in> R. \<exists>p. \<exists>(l', r') \<in> S. l = p \<bullet> l' \<and> r = p \<bullet> r')"

adhoc_overloading
  SUBSUMESEQ subsumeseq_trs

lemma subsumeseq_trsI [intro]:
  assumes "\<And>l r. (l, r) \<in> R \<Longrightarrow> \<exists>p. \<exists>(l', r') \<in> S. l = p \<bullet> l' \<and> r = p \<bullet> r'"
  shows "R \<le>\<cdot> S"
  unfolding subsumeseq_trs_def by (rule ballI2) (insert assms, auto)

lemma subsumeseq_trsE [elim]:
  assumes "R \<le>\<cdot> S" and "(l, r) \<in> R"
  obtains p l' r' where "(l', r') \<in> S" and "l = p \<bullet> l'" and "r = p \<bullet> r'"
  using assms by (auto simp: subsumeseq_trs_def)

lemma subsumeseq_trs_refl:
  fixes R :: "('f, 'v :: infinite) trs"
  shows "R \<le>\<cdot> R"
  using trs_pt.subset_imp_ex_perm [of R, OF subset_refl]
  by (auto simp: subsumeseq_trs_def split: prod.splits)
     (metis prod.inject term_pt.permute_zero)

lemma subsumeseq_trs_trans:
  fixes R S T :: "('f, 'v :: infinite) trs"
  assumes "R \<le>\<cdot> S" and "S \<le>\<cdot> T"
  shows "R \<le>\<cdot> T"
proof
  fix l r
  assume "(l, r) \<in> R"
  with \<open>R \<le>\<cdot> S\<close> obtain p l' r'
    where "(l', r') \<in> S" and "l = p \<bullet> l'" and "r = p \<bullet> r'" by blast
  moreover with \<open>S \<le>\<cdot> T\<close> obtain q l'' r''
    where "(l'', r'') \<in> T" and "l' = q \<bullet> l''" and "r' = q \<bullet> r''" by blast
  ultimately have "(l'', r'') \<in> T"
    and "l = (p + q) \<bullet> l''" and "r = (p + q) \<bullet> r''" by auto
  then show "\<exists>p. \<exists>(l', r') \<in> T. l = p \<bullet> l' \<and> r = p \<bullet> r'" by blast
qed

interpretation trs_subsumable: subsumable subsumeseq_trs
  by (default) (force simp: subsumeseq_trs_refl dest: subsumeseq_trs_trans)+

adhoc_overloading
  SUBSUMES trs_subsumable.subsumes and
  LITSIM trs_subsumable.litsim

fun num_syms :: "('f, 'v) term \<Rightarrow> nat"
where
  "num_syms (Var x) = 1" |
  "num_syms (Fun f ts) = Suc (listsum (map num_syms ts))"

fun num_funs :: "('f, 'v) term \<Rightarrow> nat"
where
  "num_funs (Var x) = 0" |
  "num_funs (Fun f ts) = Suc (listsum (map num_funs ts))"

lemma num_funs_subst:
  "num_funs (t \<cdot> \<sigma>) \<ge> num_funs t"
  by (induct t) (simp_all, metis comp_apply listsum_mono)

lemma num_syms_1: "num_syms t \<ge> 1"
  by (induct t) auto

lemma num_syms_subst:
  "num_syms (t \<cdot> \<sigma>) \<ge> num_syms t"
  using num_syms_1
  by (induct t) (auto, metis comp_apply listsum_mono)

lemma card_PCons_Collect:
  "card {i <# (p :: pos) | p. p \<in> A} = card A"
proof -
  have "{i <# p | p. p \<in> A} = (op <# i ` A)" by auto
  then show ?thesis by (auto intro: card_image simp: inj_on_def)
qed

lemma Collect_UNION:
  "{f i p | i p. i < n \<and> p \<in> A i} = (\<Union>i < n. {f i p | p. p \<in> A i})"
  by auto

lemma card_insert_poss:
  "card (insert \<epsilon> (\<Union>i < length ts. {i <# p | p. p \<in> poss (ts ! i)})) =
    Suc (card (\<Union>i < length ts. {i <# p | p. p \<in> poss (ts ! i)}))"
    (is "_ = Suc (card ?P)")
proof -
  have *: "finite ?P"
  proof -
    fix f
    have "poss (Fun f ts) \<subseteq> poss (Fun f ts)" by auto
    then show ?thesis using finite_poss [of "Fun f ts"] by (simp add: Collect_UNION)
  qed
  have "card ({\<epsilon>} \<union> ?P) = card {\<epsilon>} + card ?P"
    by (rule card_Un_disjoint) (insert *, auto)
  then show ?thesis by simp
qed

lemma card_poss_Fun:
  "card (poss (Fun f ts)) = Suc (\<Sum>i < length ts. card (poss (ts ! i)))"
proof -
  have "card (poss (Fun f ts)) = Suc (\<Sum>i < length ts. card {i <# p | p. p \<in> poss (ts ! i)})"
    by (auto simp: Collect_UNION card_insert_poss intro: card_UN_disjoint)
  then show ?thesis by (simp add: card_PCons_Collect)
qed

lemma card_poss_eq_num_syms:
  "card (poss t) = num_syms t"
  by (induct t) (simp, unfold card_poss_Fun,
                 simp add: card_poss_Fun Collect_UNION listsum_setsum_nth atLeast0LessThan)

fun num_vars :: "('f, 'v) term \<Rightarrow> nat"
where
  "num_vars (Var x) = 1" |
  "num_vars (Fun f ts) = listsum (map num_vars ts)"

definition num_unique_vars :: "('f, 'v) term \<Rightarrow> nat"
where
  "num_unique_vars t = card (vars_term t)"

lemma finite_Union_vars_term:
  "finite (\<Union>t \<in> set ts. vars_term t)"
  by auto

lemma [simp]:
  "num_unique_vars (Var x) = 1"
  by (simp_all add: num_unique_vars_def)

lemma num_unique_vars_Fun_Cons:
  "num_unique_vars (Fun f (t # ts)) \<le> num_unique_vars t + num_unique_vars (Fun f ts)"
  apply (simp_all add: num_unique_vars_def)
  unfolding card_Un_Int [OF finite_vars_term finite_Union_vars_term]
  apply simp
done

lemma listsum_map_unique_vars:
  "listsum (map num_unique_vars ts) \<ge> num_unique_vars (Fun f ts)"
proof (induct ts)
  case (Cons t ts)
  with num_unique_vars_Fun_Cons [of f t ts]
    show ?case by simp
qed (simp add: num_unique_vars_def)

lemma listsum_mono:
  fixes f g :: "'a \<Rightarrow> nat"
  assumes "\<forall>x \<in> set xs. f x \<ge> g x"
  shows "listsum (map f xs) \<ge> listsum (map g xs)"
  using assms by (induct xs) simp_all

lemma listsum_cong:
  fixes f g :: "'a \<Rightarrow> nat"
  assumes "\<forall>x \<in> set xs. f x = g x"
  shows "listsum (map f xs) = listsum (map g xs)"
  using assms by (induct xs) simp_all

lemma num_vars_geq_num_unique_vars:
  "num_vars t \<ge> num_unique_vars t"
  apply (induct t)
  apply simp_all
  using listsum_mono [of _ num_unique_vars num_vars]
    and listsum_map_unique_vars
    by (metis le_antisym le_trans linear)

lemma num_syms_geq_num_vars:
  "num_syms t \<ge> num_vars t"
proof (induct t)
  case (Fun f ts)
  with listsum_mono [of ts num_vars num_syms]
    have "listsum (map num_vars ts) \<le> listsum (map num_syms ts)" by simp
  then show ?case by simp
qed simp

lemma num_syms_eq:
  "num_syms t = num_funs t + num_vars t"
proof (induct t)
  case (Fun f ts)
  then have "\<forall>t \<in> set ts. num_syms t = num_funs t + num_vars t" by auto
  note * = listsum_cong [OF this]
    show ?case unfolding num_syms.simps *
    by (simp add: listsum_addf)
qed simp

lemma num_funs_0_imp_is_Var [simp]:
  "num_funs t = 0 \<Longrightarrow> is_Var t"
  by (induct t) auto


subsection \<open>Equality of terms modulo variables\<close>

inductive emv where
  Var [simp, intro!]: "emv (Var x) (Var y)" |
  Fun [intro]: "\<lbrakk>f = g; length ss = length ts; \<forall>i < length ts. emv (ss ! i) (ts ! i)\<rbrakk> \<Longrightarrow>
    emv (Fun f ss) (Fun g ts)"

lemma emv_refl:
  "emv t t"
  by (induct t) auto

lemma emv_sym:
  assumes "emv s t"
  shows "emv t s"
  using assms by (induct) auto

lemma emv_Var:
  assumes "emv (Var x) t"
  obtains y where "t = Var y"
  using assms by (cases) simp

lemma emv_Fun_Var [simp]:
  "emv (Fun f ts) (Var x) = False"
  by (auto elim: emv.cases)

lemma emv_Fun_Fun [iff]:
  "emv (Fun f ss) (Fun g ts) \<longleftrightarrow>
    f = g \<and> length ss = length ts \<and> (\<forall>i < length ts. emv (ss ! i) (ts ! i))"
  by (auto elim: emv.cases)

lemma emv_trans:
  assumes "emv s t" and "emv t u"
  shows "emv s u"
using assms
proof (induct arbitrary: u)
  case (Var x y)
  then show ?case by (auto elim: emv_Var)
next
  case (Fun f g ss ts)
  then show ?case by (cases u) simp_all
qed

lemma num_funs_0:
  assumes "num_funs t = 0"
  obtains x where "t = Var x"
  using assms by (induct t) auto

lemma listsum_map_num_syms_subst:
  assumes "listsum (map (num_syms \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts) = listsum (map num_syms ts)"
  shows "\<forall>i < length ts. num_syms (ts ! i \<cdot> \<sigma>) = num_syms (ts ! i)"
using assms
proof (induct ts)
  case (Cons t ts)
  then have "num_syms (t \<cdot> \<sigma>) + listsum (map (num_syms \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts)
    = num_syms t + listsum (map num_syms ts)" by (simp add: o_def)
  moreover have "num_syms (t \<cdot> \<sigma>) \<ge> num_syms t" by (metis num_syms_subst)
  moreover have "listsum (map (num_syms \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts) \<ge> listsum (map num_syms ts)"
    using num_syms_subst [of _ \<sigma>] by (induct ts) (auto intro: add_mono)
  ultimately show ?case using Cons by (auto) (case_tac i, auto)
qed simp

lemma listsum_map_num_funs_subst:
  assumes "listsum (map (num_funs \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts) = listsum (map num_funs ts)"
  shows "\<forall>i < length ts. num_funs (ts ! i \<cdot> \<sigma>) = num_funs (ts ! i)"
using assms
proof (induct ts)
  case (Cons t ts)
  then have "num_funs (t \<cdot> \<sigma>) + listsum (map (num_funs \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts)
    = num_funs t + listsum (map num_funs ts)" by (simp add: o_def)
  moreover have "num_funs (t \<cdot> \<sigma>) \<ge> num_funs t" by (metis num_funs_subst)
  moreover have "listsum (map (num_funs \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts) \<ge> listsum (map num_funs ts)"
    using num_funs_subst [of _ \<sigma>] by (induct ts) (auto intro: add_mono)
  ultimately show ?case using Cons by (auto) (case_tac i, auto)
qed simp

lemma subst_size_emv:
  assumes "s = t \<cdot> \<tau>" and "num_syms s = num_syms t" and "num_funs s = num_funs t"
  shows "emv s t"
using assms
proof (induct t arbitrary: s)
  case (Var x)
  then show ?case by (force elim: num_funs_0)
next
  case (Fun g ts)
  note IH = this
  show ?case
  proof (cases s)
    case (Var x)
    then show ?thesis using Fun by simp
  next
    case (Fun f ss)
    from IH(2-) [unfolded Fun]
      and listsum_map_num_syms_subst [of \<tau> ts]
      and listsum_map_num_funs_subst [of \<tau> ts]
      have "\<forall>i < length ts. num_syms (ts ! i  \<cdot> \<tau>) = num_syms (ts ! i)"
        and "\<forall>i < length ts. num_funs (ts ! i \<cdot> \<tau>) = num_funs (ts ! i)"
        by auto
    with Fun and IH show ?thesis by auto
  qed
qed

lemma subsumeseq_term_size_emv:
  assumes "s \<cdot>\<ge> t" and "num_syms s = num_syms t" and "num_funs s = num_funs t"
  shows "emv s t"
  using assms(1) and subst_size_emv [OF _ assms(2-)] by (cases) simp

lemma emv_subst_vars_term:
  assumes "emv s t"
    and "s = t \<cdot> \<sigma>"
  shows "vars_term s = (the_Var \<circ> \<sigma>) ` vars_term t"
using assms [unfolded subsumeseq_term_iff]
apply (induct)
apply (auto simp: in_set_conv_nth iff: image_iff)
apply (metis nth_mem)
by (metis comp_apply imageI nth_mem)

lemma emv_subst_imp_num_unique_vars_le:
  assumes "emv s t"
    and "s = t \<cdot> \<sigma>"
  shows "num_unique_vars s \<le> num_unique_vars t"
  using emv_subst_vars_term [OF assms]
  apply (simp add: num_unique_vars_def)
  by (metis card_image_le finite_vars_term)

lemma emv_subsumeseq_term_imp_num_unique_vars_le:
  assumes "emv s t"
    and "s \<cdot>\<ge> t"
  shows "num_unique_vars s \<le> num_unique_vars t"
  using assms(2) and emv_subst_imp_num_unique_vars_le [OF assms(1)] by (cases) simp

lemma num_syms_ge_num_unique_vars:
  "num_syms t \<ge> num_unique_vars t"
  by (metis le_trans num_syms_geq_num_vars num_vars_geq_num_unique_vars)

lemma inc_seq_greater:
  fixes f :: "nat seq"
  assumes "\<forall>i. f i < f (Suc i)"
  shows "\<exists>i. f i > N"
  using assms
  apply (induct N)
  apply (auto)
  apply (metis neq0_conv)
  by (metis Suc_lessI)

lemma num_syms_num_unique_vars_clash:
  assumes "\<forall>i. num_syms (f i) = num_syms (f (Suc i))"
    and "\<forall>i. num_unique_vars (f i) < num_unique_vars (f (Suc i))"
  shows False
proof -
  have *: "\<forall>i j. i \<le> j \<longrightarrow> num_syms (f i) = num_syms (f j)"
  proof (intro allI impI)
    fix i j :: nat
    assume "i \<le> j"
    then show "num_syms (f i) = num_syms (f j)"
      using assms(1)
      apply (induct "j - i" arbitrary: i)
      apply auto
      by (metis Suc_diff_diff diff_zero less_eq_Suc_le order.not_eq_order_implies_strict)
  qed
  have "\<exists>i. num_unique_vars (f i) \<ge> num_syms (f 0)"
    using inc_seq_greater [OF assms(2), of "num_syms (f 0)"] by (metis nat_less_le)
  then obtain i where "num_unique_vars (f i) \<ge> num_syms (f 0)" by auto
  with * and assms(2) have "num_unique_vars (f (Suc i)) > num_syms (f (Suc i))"
    by (metis le0 le_antisym num_syms_ge_num_unique_vars)
  then show False
    by (metis less_Suc_eq_le not_less_eq num_syms_ge_num_unique_vars)
qed

lemma bij_Var_subst_compose_Var:
  assumes "bij g"
  shows "(Var \<circ> g) \<circ>\<^sub>s (Var \<circ> inv g) = Var"
proof
  fix x
  show "((Var \<circ> g) \<circ>\<^sub>s (Var \<circ> inv g)) x = Var x"
    using assms
    apply (auto simp: subst_compose_def)
    by (metis UNIV_I bij_is_inj inv_into_f_f)
qed

lemma emv_subst_imp_is_Var:
  assumes "emv s t"
    and "s = t \<cdot> \<sigma>"
  shows "\<forall>x \<in> vars_term t. is_Var (\<sigma> x)"
  using assms
  apply (induct)
  apply auto
  by (metis in_set_conv_nth)

lemma wf_subsumes:
  "wf ({<\<cdot>} :: ('f, 'v) term rel)"
proof (rule ccontr)
  assume "\<not> ?thesis"
  then obtain f :: "('f, 'v) term seq"
    where strict: "\<forall>i. f i \<cdot>> f (Suc i)"
    by (metis mem_Collect_eq splitD wf_iff_no_infinite_down_chain)
  then have *: "\<forall>i. f i \<cdot>\<ge> f (Suc i)" by (metis term_subsumable.subsumption.less_imp_le)
  then have "\<forall>i. num_syms (f i) \<ge> num_syms (f (Suc i))"
    by (auto simp: subsumeseq_term_iff) (metis num_syms_subst)
  from down_chain_imp_eq [OF this] obtain N
    where N_syms: "\<forall>i > N. num_syms (f i) = num_syms (f (Suc i))" ..
  def g \<equiv> "\<lambda>i. f (i + N)"
  from * have "\<forall>i. num_funs (g i) \<ge> num_funs (g (Suc i))"
    by (auto simp: subsumeseq_term_iff g_def) (metis num_funs_subst)
  from down_chain_imp_eq [OF this] obtain K
    where K_funs: "\<forall>i > K. num_funs (g i) = num_funs (g (Suc i))" ..
  def M \<equiv> "max K N"
  have strict_g: "\<forall>i > M. g i \<cdot>> g (Suc i)" using strict by (simp add: g_def M_def)
  have g: "\<forall>i > M. g i \<cdot>\<ge> g (Suc i)" using * by (simp add: g_def M_def)
  moreover have "\<forall>i > M. num_funs (g i) = num_funs (g (Suc i))"
    using K_funs unfolding M_def by (metis max_less_iff_conj)
  moreover have syms: "\<forall>i > M. num_syms (g i) = num_syms (g (Suc i))"
    using N_syms unfolding M_def g_def
    by (metis add_Suc_right add_lessD1 add_strict_left_mono add.commute)
  ultimately have emv: "\<forall>i > M. emv (g i) (g (Suc i))" by (metis subsumeseq_term_size_emv)
  then have "\<forall>i > M. num_unique_vars (g (Suc i)) \<ge> num_unique_vars (g i)"
    using emv_subsumeseq_term_imp_num_unique_vars_le and g by fast
  then obtain i where "i > M"
    and nuv: "num_unique_vars (g (Suc i)) = num_unique_vars (g i)"
    using num_syms_num_unique_vars_clash [of "\<lambda>i. g (i + Suc M)"] and syms
    by (metis add_Suc_right add_Suc_shift le_eq_less_or_eq less_add_Suc2)
  def s \<equiv> "g i" and t \<equiv> "g (Suc i)"
  from nuv have card: "card (vars_term s) = card (vars_term t)"
    by (simp add: num_unique_vars_def s_def t_def)
  from g [THEN spec, THEN mp, OF \<open>i > M\<close>] obtain \<sigma>
    where "s = t \<cdot> \<sigma>" by (cases) (auto simp: s_def t_def)
  then have "emv s t" and "vars_term s = (the_Var \<circ> \<sigma>) ` vars_term t"
    using emv_subst_vars_term [of s t \<sigma>] and emv and \<open>i > M\<close> by (auto simp: s_def t_def)
  with card have "card ((the_Var \<circ> \<sigma>) ` vars_term t) = card (vars_term t)" by simp
  from finite_card_eq_imp_bij_betw [OF finite_vars_term this]
    have "bij_betw (the_Var \<circ> \<sigma>) (vars_term t) ((the_Var \<circ> \<sigma>) ` vars_term t)" .

  from bij_betw_extend [OF this, of UNIV]
    obtain h where *: "\<forall>x\<in>vars_term t. h x = (the_Var \<circ> \<sigma>) x"
    and "finite {x. h x \<noteq> x}"
    and "bij h" by auto
  then have perm: "h \<in> perms" by (auto simp: perms_def)
  have "\<forall>x\<in>vars_term t. (Var \<circ> h) x = \<sigma> x"
  proof
    fix x
    assume "x \<in> vars_term t"
    with * have "h x = (the_Var \<circ> \<sigma>) x" by simp
    with emv_subst_imp_is_Var [OF \<open>emv s t\<close> \<open>s = t \<cdot> \<sigma>\<close>] \<open>x \<in> vars_term t\<close>
      show "(Var \<circ> h) x = \<sigma> x" by simp
  qed
  then have "t \<cdot> (Var \<circ> h) = s"
    using \<open>s = t \<cdot> \<sigma>\<close> by (auto simp: term_subst_eq_conv)
  then have "t \<cdot> (Var \<circ> h) \<circ>\<^sub>s (Var \<circ> inv h) = s \<cdot> (Var \<circ> inv h)" by auto
  then have "t = s \<cdot> (Var \<circ> inv h)"
    unfolding bij_Var_subst_compose_Var [OF \<open>bij h\<close>] by simp
  then have "t \<cdot>\<ge> s" by auto
  with strict_g and \<open>i > M\<close> show False by (auto simp: s_def t_def term_subsumable.subsumes_def)
qed


subsection \<open>The Encompassment Relation\<close>

inductive encompeq :: "('f, 'v) term \<Rightarrow> ('f, 'v) term \<Rightarrow> bool" (infix "\<unlhd>\<cdot>" 50)
where
  [intro]: "t = C\<langle>s \<cdot> \<sigma>\<rangle> \<Longrightarrow> s \<unlhd>\<cdot> t"

lemma encompeq_termE [elim]:
  assumes "s \<unlhd>\<cdot> t"
  obtains C \<sigma> where "t = C\<langle>s \<cdot> \<sigma>\<rangle>"
  using assms by (cases)

lemma encompeq_refl:
  "t \<unlhd>\<cdot> t"
  by (rule encompeq.intros [of t \<box> t Var]) simp

lemma encompeq_trans:
  assumes "s \<unlhd>\<cdot> t"
    and "t \<unlhd>\<cdot> u"
  shows "s \<unlhd>\<cdot> u"
proof -
  obtain C D \<sigma> \<tau>
    where [simp]: "t = C\<langle>s \<cdot> \<sigma>\<rangle>" "u = D\<langle>t \<cdot> \<tau>\<rangle>" using assms by fastforce
  show ?thesis
    by (rule encompeq.intros [of _ "D \<circ>\<^sub>c (C \<cdot>\<^sub>c \<tau>)" _ "\<sigma> \<circ>\<^sub>s \<tau>"]) simp
qed

definition encomp (infix "\<lhd>\<cdot>" 50)
where
  "t \<lhd>\<cdot> s \<longleftrightarrow> t \<unlhd>\<cdot> s \<and> \<not> s \<unlhd>\<cdot> t"

abbreviation (input) revencompeq  (infix "\<cdot>\<unrhd>" 50) where
  "x \<cdot>\<unrhd> y \<equiv> y \<unlhd>\<cdot> x"

abbreviation (input) revencomp (infix "\<cdot>\<rhd>" 50) where
  "x \<cdot>\<rhd> y \<equiv> y \<lhd>\<cdot> x"

abbreviation encompeq_set ("{\<unlhd>\<cdot>}")
where
  "{\<unlhd>\<cdot>} \<equiv> {(x, y). x \<unlhd>\<cdot> y}"

abbreviation encomp_set ("{\<lhd>\<cdot>}")
where
  "{\<lhd>\<cdot>} \<equiv> {(x, y). x \<lhd>\<cdot> y}"

abbreviation revencompeq_set ("{\<cdot>\<unrhd>}")
where
  "{\<cdot>\<unrhd>} \<equiv> {(x, y). y \<unlhd>\<cdot> x}"

abbreviation revencomp_set ("{\<cdot>\<rhd>}")
where
  "{\<cdot>\<rhd>} \<equiv> {(x, y). y \<lhd>\<cdot> x}"

interpretation encomp_order: preorder "op \<unlhd>\<cdot>" "op \<lhd>\<cdot>"
  by (default) (auto simp: encomp_def encompeq_refl elim: encompeq_trans)

text \<open>Encompassment is the composition of subsumption and the subterm relation.\<close>
lemma encompeq_eq_subteq_instance:
  "s \<unlhd>\<cdot> t \<longleftrightarrow> (op \<le>\<cdot> OO op \<unlhd>) s t"
  by (force simp: OO_def subsumeseq_term_iff)

lemma subsumeseq_size:
  "s \<le>\<cdot> t \<Longrightarrow> size s \<le> size t"
  by (cases s t rule: subsumeseq_term.cases) (auto simp: size_subst)

lemma subsumeseq_imp_encompeq:
  "s \<le>\<cdot> t \<Longrightarrow> s \<unlhd>\<cdot> t"
  by (cases s t rule: subsumeseq_term.cases) (auto intro: encompeq.intros [of _ \<box>])

lemma encompeq_size:
  "s \<unlhd>\<cdot> t \<Longrightarrow> size s \<le> size t"
  by (force simp: encompeq_eq_subteq_instance dest: subsumeseq_size supteq_size)

lemma encompeq_subsumeseq_cases:
  "s \<unlhd>\<cdot> t \<longleftrightarrow> s \<le>\<cdot> t \<or> (op \<le>\<cdot> OO op \<lhd>) s t"
  by (auto simp: encompeq_eq_subteq_instance)

lemma encomp_subsumes_cases:
  "s \<lhd>\<cdot> t \<longleftrightarrow> s <\<cdot> t \<or> (op \<le>\<cdot> OO op \<lhd>) s t"
  by (auto simp: encomp_def term_subsumable.subsumes_def encompeq_subsumeseq_cases OO_def)
     (auto dest!: supt_size subsumeseq_size)

lemma wf_subsumeseq_O_subt:
  "wf ({\<le>\<cdot>} O {(x, y). x \<lhd> y})"
proof -
  have "\<And>x y. (x, y) \<in> {\<le>\<cdot>} O {(x, y). x \<lhd> y} \<Longrightarrow> size x < size y"
    by (auto simp: OO_def dest!: subsumeseq_size supt_size)
  then have "{\<le>\<cdot>} O {(x, y). x \<lhd> y} \<subseteq> inv_image {(x, y). x < y} size"
    by (auto simp: inv_image_def)
  moreover have "wf (inv_image {(x, y). x < y} size)" by (rule wf_inv_image [OF wf_less])
  ultimately show ?thesis by (metis wf_subset)
qed

lemma litsim_encompeq_iff:
  "s \<doteq> t \<longleftrightarrow> s \<unlhd>\<cdot> t \<and> t \<unlhd>\<cdot> s"
proof
  assume "s \<doteq> t"
  then show "s \<unlhd>\<cdot> t \<and> t \<unlhd>\<cdot> s"
    by (auto simp: term_subsumable.litsim_def dest: subsumeseq_imp_encompeq)
next
  presume "s \<unlhd>\<cdot> t" and "t \<unlhd>\<cdot> s"
  then have "s \<le>\<cdot> t" and "t \<le>\<cdot> s"
    by (auto simp: encompeq_subsumeseq_cases dest!: supt_size subsumeseq_size)
  then show "s \<doteq> t" by (auto simp: term_subsumable.litsim_def)
qed auto

lemma encompeq_litsim_encomp_iff:
  "s \<unlhd>\<cdot> t \<longleftrightarrow> s \<doteq> t \<or> s \<lhd>\<cdot> t"
  by (auto simp: litsim_encompeq_iff encomp_def)

lemmas encompeq_cases [consumes 1] = encompeq_litsim_encomp_iff [THEN iffD1]

lemma encomp_supt:
  assumes "s \<cdot>> t" and "t \<rhd> u"
  shows "\<exists>t \<lhd> s. t \<cdot>\<ge> u"
proof -
  from assms [unfolded term_subsumable.subsumes_def subsumeseq_term_iff] obtain \<sigma>
    where "s = t \<cdot> \<sigma>" and "\<not> (\<exists>\<sigma>. t = s \<cdot> \<sigma>)" by blast 
  moreover with \<open>t \<rhd> u\<close> have "t \<cdot> \<sigma> \<rhd> u \<cdot> \<sigma>"
    by (metis instance_no_supt_imp_no_supt)
  ultimately have "s \<rhd> u \<cdot> \<sigma>" and "u \<cdot> \<sigma> \<cdot>\<ge> u"
    by (auto simp: term_subsumable.subsumes_def)
  then show ?thesis by blast
qed

lemma wf_encomp:
  "wf ({\<lhd>\<cdot>} :: ('f, 'v) term rel)"
proof -
  have "{\<rhd>} O {\<cdot>\<ge>} \<union> {\<cdot>>} = ({\<cdot>\<rhd>} :: ('f, 'v) term rel)" by (auto simp: encomp_subsumes_cases)
  moreover
  have "SN ({\<rhd>} O {\<cdot>\<ge>} \<union> ({\<cdot>>} :: ('f, 'v) term rel))"
  proof (rule quasi_commute_imp_SN)
    have [simp]: "{(s, u). \<exists>t. s \<le>\<cdot> t \<and> u \<rhd> t} = {(s, u). \<exists>t \<lhd> u. s \<le>\<cdot> t}" by auto
    show "SN ({\<rhd>} O {\<cdot>\<ge>})"
      using wf_subsumeseq_O_subt
      by (auto simp: SN_iff_wf converse_unfold OO_def relcomp_def)
    show "SN ({\<cdot>>} :: ('f, 'v) term rel)"
      using wf_subsumes by (auto simp: SN_iff_wf converse_unfold)
    show "quasi_commute ({\<rhd>} O {\<cdot>\<ge>}) ({\<cdot>>} :: ('f, 'v) term rel)"
    proof (unfold quasi_commute_def, intro subrelI)
      fix s t :: "('f, 'v) term"
      assume "(s, t) \<in> {\<cdot>>} O {\<rhd>} O {\<cdot>\<ge>}"
      moreover have "{\<cdot>>} O {\<rhd>} \<subseteq> {\<rhd>} O {\<cdot>\<ge>}"
        by (auto simp: relcomp_def OO_def dest: encomp_supt)
      ultimately have "(s, t) \<in> {\<rhd>} O {\<cdot>\<ge>} O {\<cdot>\<ge>}" by auto
      then have "(s, t) \<in> {\<rhd>} O {\<cdot>\<ge>}"
        by (auto simp: relcomp_def OO_def dest: subsumeseq_term_trans)
      then show "(s, t) \<in> ({\<rhd>} O {\<cdot>\<ge>}) O ({\<rhd>} O {\<cdot>\<ge>} \<union> {\<cdot>>})\<^sup>*" by auto
    qed
  qed
  ultimately
  show ?thesis by (auto simp: SN_iff_wf converse_unfold)
qed

lemmas encomp_induct = wfP_induct [OF wf_encomp [to_pred]]

end

