(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2012-2015)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2012-2015)
License: LGPL (see file COPYING.LESSER)
*)
theory KBO_Impl
imports
  KBO
  Term_Order_Impl
  "../Rewriting/Term_Impl"
  "../Auxiliaries/Multiset_Code"
  "../Framework/Map_Choice"
begin

(* in the implementation, the precedence is determined by a mapping into naturals (0 has least prec.) *)
locale weight_fun_nat_prc =
  fixes w :: "'f \<times> nat \<Rightarrow> nat"
  and w0 :: nat
  and prc :: "'f \<times> nat \<Rightarrow> nat"
  and least :: "'f \<Rightarrow> bool"
  and scf :: "'f \<times> nat \<Rightarrow> nat \<Rightarrow> nat"

sublocale weight_fun_nat_prc \<subseteq> weight_fun_prc w w0 scf least "\<lambda> fn gm. prc fn > prc gm" "\<lambda> fn gm. prc fn \<ge> prc gm" .

context weight_fun_nat_prc
begin

definition kbo_impl where "kbo_impl \<equiv> kbo"

(* main difference to kbo from KBO.thy:
  - executable
  - in Var-Var case, there is no check on equal variables to slightly improve efficiency.
*)
lemma kbo_impl: 
  "kbo_impl s t = (let wt = weight t; ws = weight s in if (vars_term_ms (SCF t) \<le> vars_term_ms (SCF s) \<and> wt \<le> ws)
      then if (wt < ws) then (True,True)
      else (case s of
      Var y \<Rightarrow> (False, (case t of Var x \<Rightarrow> True | Fun g ts \<Rightarrow> ts = [] \<and> least g)) 
    | Fun f ss \<Rightarrow> case t of Var x \<Rightarrow> (True,True)
                 | Fun g ts \<Rightarrow> let pf = prc (f,length ss); pg = prc (g,length ts) in if pf > pg then (True,True) else if pf \<ge> pg then 
              lex_ext_unbounded kbo_impl ss ts else (False,False))
       else (False,False))" 
proof -
  {
    fix x y
    assume s: "s = Var x" and t: "t = Var y" and vars: "vars_term_ms (SCF t) \<le> vars_term_ms (SCF s)"
    from vars[unfolded s t] have "{#y#} \<le> {#x#}" by simp
    from mset_leD[OF this, of y, simplified] have "y = x" by (cases "x = y") auto
  }
  note main = this
  show ?thesis
    unfolding kbo_impl_def kbo.simps Let_def 
    by (rule if_cong[OF refl if_cong[OF refl refl term.case_cong[OF refl _ refl]] refl], insert main, cases t, auto)
qed
end

declare weight_fun_nat_prc.kbo_impl_def[simp]

abbreviation kbo where "kbo \<equiv> weight_fun_nat_prc.kbo_impl"
declare weight_fun.weight.simps[code]
declare weight_fun_nat_prc.kbo_impl[code]


(* for representing concrete instances of KBO, we define precedence, ... for finitely many
   symbols, the remainder obtains default values *)
type_synonym 'f prec_weight_repr = "(('f \<times> nat) \<times> (nat \<times> nat \<times> (nat list option))) list \<times> nat"

fun shows_kbo_repr :: "('f\<Colon>show) prec_weight_repr \<Rightarrow> shows"
where
  "shows_kbo_repr (prs, w0) =
    shows ''KBO with the following precedence and weight function'' \<circ> shows_nl \<circ>
    foldr (\<lambda>((f, n),(pr, w, scf)).
      shows ''precedence('' \<circ> shows f \<circ> shows ''['' \<circ> shows n \<circ> shows '']) = '' \<circ> shows pr \<circ> shows_nl) prs \<circ> shows_nl \<circ>
    shows ''precedence(_) = 0'' \<circ> shows_nl \<circ>
    shows ''and the following weight'' \<circ> shows_nl \<circ>
    foldr (\<lambda>((f, n), (pr, w, scf)).
      shows ''weight('' \<circ> shows f \<circ> shows ''['' \<circ> shows n \<circ> shows '']) = '' \<circ> shows w \<circ> shows_nl) prs \<circ> shows_nl \<circ>
    shows ''weight(_) = '' \<circ> shows (Suc w0) \<circ> shows_nl \<circ>
    shows ''w0 = '' \<circ> shows w0 \<circ> shows_nl \<circ>
    shows ''and the following subterm coefficient functions'' \<circ> shows_nl \<circ>
    foldr (\<lambda>((f, n), (pr, w, scf)).
      shows ''scf('' \<circ> shows f \<circ> shows ''['' \<circ> shows n \<circ> shows '']) = '' \<circ>
      (if scf = None then shows ''all 1'' else shows (the scf)) \<circ> shows_nl) prs \<circ> shows_nl \<circ>
    shows ''scf(_) = all 1'' \<circ> shows_nl"

definition scf_repr_to_scf :: "('f \<times> nat \<Rightarrow> nat list option) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat \<Rightarrow> nat)" where
  "scf_repr_to_scf scf fn i \<equiv> case scf fn of None \<Rightarrow> 1 | Some xs \<Rightarrow> xs ! i"

fun check_scf_entry :: "('f :: show) \<times> nat \<Rightarrow> nat list option \<Rightarrow> shows check" where
  "check_scf_entry fn None = succeed"
| "check_scf_entry (f,n) (Some es) = do {
     check (length es = n) (shows ''nr of entries should be '' +@+ shows n);
     check (\<forall> e \<in> set es. e > 0) (shows ''all entries must be non-zero'')
   } <+? (\<lambda> s. shows ''problem with subterm coefficients for '' +@+ shows (f,n) +@+ shows '': '' +@+ s +@+ shows_nl)"

(* check whether partially defined prec and weight are admissible, convert to total weight and prec *)
definition
  prec_weight_repr_to_prec_weight :: "('f :: {show,key}) prec_weight_repr \<Rightarrow> shows check \<times> ('f \<times> nat \<Rightarrow> nat) \<times> ('f \<times> nat \<Rightarrow> nat) \<times> nat \<times> ('f \<Rightarrow> bool) \<times> ('f \<times> nat \<Rightarrow> nat \<Rightarrow> nat)" 
where
  "prec_weight_repr_to_prec_weight prw_w0 \<equiv>
    let 
      (prw, w0) = prw_w0;
      prwm    = ceta_map_of prw;
      w_fun   = fun_of_map_fun' prwm (\<lambda> _. Suc w0) (fst o snd);
      p_fun   = fun_of_map_fun' prwm (\<lambda> _. 0) fst;
      scf_fun = fun_of_map_fun' prwm (\<lambda> _. None) (snd o snd);
      fs      = map fst prw;
      p_max   = max_list (map p_fun fs);
      cw_okay = check_allm (\<lambda> fn. check (snd fn = 0 \<longrightarrow> w_fun fn \<ge> w0) (shows ''weight of constant '' +@+ shows (fst fn) +@+ shows '' must be at least w0'')) (map fst prw);
      adm     = check_allm (\<lambda> fn. check (snd fn = 1 \<longrightarrow> w_fun fn = 0 \<longrightarrow> p_fun fn = p_max) (shows ''unary symbol '' +@+ shows (fst fn) +@+ shows '' with weight 0 does not have maximal precedence'')) (map fst prw);
      scf_ok  = check_allm (\<lambda> fn. check_scf_entry fn (scf_fun fn)) (map fst prw);
      cs      = map fst (filter (\<lambda> fn. snd fn = 0 \<and> w_fun fn = w0) fs);
      p_min   = min_list (map (\<lambda> c. p_fun (c,0)) cs);
      lcs     = filter (\<lambda> c. p_fun (c,0) = p_min) cs;
      least   = (\<lambda> c. c \<in> set lcs);
      ok      = (do {
        check (w0 > 0) (shows ''w0 must be larger than 0'');
        adm;
        cw_okay;
        scf_ok
      })
      in (ok, p_fun, w_fun, w0, least, scf_repr_to_scf scf_fun)"

lemma prec_weight_repr_to_prec_weight: 
  assumes ok: "prec_weight_repr_to_prec_weight prw_w0 = (succeed,p,w,w0,least,scf_fun)"
  shows "admissible_weight_fun_prc w w0 (\<lambda> fn gm. p fn > p gm) (\<lambda> fn gm. p fn \<ge> p gm) least scf_fun"
proof -
  have id: "\<And> b. (b = succeed) = isOK(b)" by auto
  obtain prw w0' where prw_w0: "prw_w0 = (prw,w0')" by force
  note ok = ok[unfolded prec_weight_repr_to_prec_weight_def prw_w0 Let_def split]
  from ok have w0': "w0' = w0" by simp
  note ok = ok[unfolded w0']
  obtain fs where fs: "fs = set (map fst prw)" by auto
  note ok = ok[simplified, unfolded id, simplified]
  have w: "w = (\<lambda> fn. case map_of prw fn of Some pw \<Rightarrow> (fst o snd) pw | None \<Rightarrow> Suc w0)" 
    by (rule ext, insert ok, auto)
  have p: "p = (\<lambda> fn. case map_of prw fn of Some pw \<Rightarrow> fst pw | None \<Rightarrow> 0)"
    by (rule ext, insert ok, auto)
  let ?scf = "(\<lambda> fn. case map_of prw fn of Some pw \<Rightarrow> (snd o snd) pw | None \<Rightarrow> None)"
  have scf: "scf_fun = scf_repr_to_scf ?scf"
    by (rule ext, insert ok, auto simp: scf_repr_to_scf_def)
  from ok have cw_okay: "\<And> fn. fn \<in> fs \<Longrightarrow> snd fn = 0 \<Longrightarrow> w fn \<ge> w0" unfolding w fs by auto
  obtain pmax where pmax: "pmax = max_list (map p (map fst prw))" by auto
  from ok have adm: "\<And> fn. fn \<in> fs \<Longrightarrow> snd fn = 1 \<Longrightarrow> w fn = 0 \<Longrightarrow> p fn = pmax"
    unfolding pmax p w fs by auto
  from ok have w0: "w0 > 0" by simp
  obtain cs where cs: "cs = map fst (filter (\<lambda> fn. snd fn = 0 \<and> w fn = w0) (map fst prw))" by auto
  obtain pmin where pmin: "pmin = min_list (map (\<lambda> c. p (c,0)) cs)" by auto
  obtain lcs where lcs: "lcs = filter (\<lambda> c. p (c,0) = pmin) cs" by auto
  have least: "least = (\<lambda> c. c \<in> set lcs)"
    using ok unfolding lcs pmin cs fs by auto
  {
    fix fn
    assume "fn \<notin> fs"
    hence fn: "\<And> e. (fn,e) \<notin> set prw" unfolding fs by force
    with map_of_SomeD[of prw fn] have l: "map_of prw fn = None" 
      by (cases "map_of prw fn", auto)
    hence "p fn = 0 \<and> w fn = Suc w0 \<and> ?scf fn = None" unfolding p w by auto
  } note not_fs = this
  show ?thesis
  proof
    show "w0 > 0" by (rule w0)
  next    
    fix f
    show "w0 \<le> w (f,0)"
    proof (cases "(f,0) \<in> fs")
      case False
      from not_fs[OF this] show ?thesis by simp
    next
      case True
      from cw_okay[OF this] show ?thesis by simp
    qed
  next
    fix f g n
    assume wf: "w (f,1) = 0"
    show "p (g,n) \<le> p (f,1)" 
    proof (cases "(g,n) \<in> fs")
      case False
      from not_fs[OF this] show ?thesis by simp
    next
      case True
      from not_fs[of "(f,1)"] wf have f: "(f,1) \<in> fs" by auto
      from True have "p (g,n) \<in> set (map p (map fst prw))" unfolding fs by force
      from max_list[OF this] have "p (g,n) \<le> pmax" unfolding pmax .
      also have "... = p (f,1)" using adm[OF f _ wf] by simp
      finally show "p (g,n) \<le> p (f,1)" by simp
    qed
  next
    fix f
    show "least f = (w (f,0) = w0 \<and> (\<forall> g. w (g,0) = w0 \<longrightarrow> p (f,0) \<le> p (g,0)))" (is "_ = ?r")
    proof (cases "(f,0) \<in> fs \<and> w (f,0) = w0")
      case False
      hence "f \<notin> set cs" unfolding cs fs by force
      hence "least f = False" unfolding least lcs by simp
      also have "... = ?r"
      proof (cases "(f,0) \<in> fs")
        case True
        thus ?thesis using False by simp
      next
        case False
        thus ?thesis using not_fs[OF False] by simp
      qed
      finally show ?thesis .
    next
      case True
      let ?ps = "map (\<lambda> c. p (c,0)) cs"
      {
        fix c
        assume wc: "w (c,0) = w0"
        with not_fs[of "(c,0)"] have c: "(c,0) \<in> fs" by auto
        with wc have c: "c \<in> set cs" unfolding cs fs by force
        hence "p (c,0) \<in> set ?ps" by auto
        from pmin min_list[OF this] have "pmin \<le> p (c,0)" by simp
      } note pmin_c = this        
      from True have f: "f \<in> set cs" and wf: "w (f,0) = w0" unfolding cs fs by force+      
      show ?thesis 
      proof (cases "f \<in> set lcs")
        case True
        hence least: "least f = True" unfolding least by simp
        have ?r
        proof (rule conjI[OF wf], intro allI impI)
          fix g
          assume wg: "w (g,0) = w0"
          from pmin_c[OF this] have pg: "pmin \<le> p (g,0)" by simp
          from True[unfolded lcs] have pf: "p (f,0) = pmin" by simp
          from pf pg show "p (f,0) \<le> p (g,0)" by simp
        qed
        with least show ?thesis by simp
      next
        case False
        from False[unfolded lcs] f have "p (f,0) \<noteq> pmin" by force
        with pmin_c[OF wf] have pf: "p (f,0) > pmin" by auto
        from f have "?ps \<noteq> []" by auto
        from min_list_ex[OF this] obtain pm where pm: "pm \<in> set ?ps" 
          and pmin: "pm = pmin" unfolding pmin by auto
        hence pmin: "pmin \<in> set ?ps" by auto
        then obtain c where c: "c \<in> set cs" and pc: "p (c,0) = pmin" by auto
        from c[unfolded cs] have wc: "w (c,0) = w0" by auto
        from pf pc have p: "p (c,0) < p (f,0)" by simp
        from False p wc  show ?thesis unfolding least by auto
      qed
    qed
  next
    fix f and n i :: nat
    assume i: "i < n"
    def scffn \<equiv> "?scf (f,n)"
    show "0 < scf_fun (f,n) i"
    proof (cases "map_of prw (f, n)")
      case None
      thus ?thesis unfolding scf scf_repr_to_scf_def by auto
    next
      case (Some e)
      def s \<equiv> "(snd o snd) e"
      from map_of_SomeD[OF Some]
      have "((f,n),e) \<in> set prw" by force
      hence "isOK(check_scf_entry (f,n) (?scf (f,n)))"
        using ok by auto
      also have "?scf (f,n) = s" unfolding s_def Some by simp
      finally have ok: "isOK(check_scf_entry (f,n) s)" .
      show ?thesis unfolding scf scf_repr_to_scf_def Some option.simps s_def[symmetric]
        using ok i by (cases s, auto)
    qed
  qed (insert SN_inv_image[OF SN_nat_gt, of p], auto simp: inv_image_def)
qed

(* checking kbo constraints: invoke kbo, and throw error message if terms are not in relation *)                  
definition
  kbo_strict :: "('f \<times> nat \<Rightarrow> nat) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f :: show,'v :: show)rule \<Rightarrow> shows check"
where
  "kbo_strict pr w w0 least scf \<equiv> \<lambda>(s, t). check (fst (kbo w w0 pr least scf s t)) (''could not orient '' +#+ shows s +@+ '' >KBO '' +#+ shows t +@+ shows_nl)"

lemma kbo_strict[simp]: "isOK (kbo_strict pr w w0 least scf st) = fst (kbo w w0 pr least scf (fst st) (snd st))" unfolding kbo_strict_def by (cases st, auto)


definition
  kbo_nstrict :: "('f \<times> nat \<Rightarrow> nat) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat) \<Rightarrow> nat \<Rightarrow> ('f \<Rightarrow> bool) \<Rightarrow> ('f \<times> nat \<Rightarrow> nat \<Rightarrow> nat) \<Rightarrow> ('f :: show,'v :: show)rule \<Rightarrow> shows check"
where
  "kbo_nstrict pr w w0 least scf \<equiv> \<lambda>(s, t). check (snd (kbo w w0 pr least scf s t)) (''could not orient '' +#+ shows s +@+ '' >=KBO '' +#+ shows t +@+ shows_nl)"

lemma kbo_nstrict[simp]: "isOK (kbo_nstrict pr w w0 least scf st) = snd (kbo w w0 pr least scf (fst st) (snd st))" unfolding kbo_nstrict_def by (cases st, auto)

fun create_KBO_redtriple :: "(('f :: show)prec_weight_repr \<Rightarrow> 'g prec_weight_repr) \<Rightarrow> 'f prec_weight_repr \<Rightarrow> ('g :: {show,key},'v :: show)redtriple"
where "create_KBO_redtriple f_to_g pr = (let (ch,p,w,w0,least,scf) = prec_weight_repr_to_prec_weight (f_to_g pr);
   ns = kbo_nstrict p w w0 least scf;
   s =  kbo_strict p w w0 least scf
    in
  \<lparr> redtriple.valid = ch, 
    s = s, 
    ns = ns, 
    nst = ns, 
    af = full_af, 
    mono_af = full_af,
    mono = (\<lambda> _. succeed), 
    desc = shows_kbo_repr pr, 
    not_ws_ns = Some [],
    cpx = no_complexity\<rparr>)"

(* register KBO as executable reduction triple *)
interpretation KBO_redpair: generic_redtriple_impl "create_KBO_redtriple (f_to_g :: ('f :: show)prec_weight_repr \<Rightarrow> ('g :: {show,key})prec_weight_repr)"
proof
  fix "prw" :: "'f prec_weight_repr"
   and s_list ns_list nst_list :: "('g,'v :: show)rule list"
  let ?rp = "create_KBO_redtriple f_to_g prw :: ('g,'v)redtriple"
  let ?af = "redtriple.af ?rp :: ('g af)"
  let ?af' = "redtriple.mono_af ?rp :: ('g af)"
  let ?pr = "prec_weight_repr_to_prec_weight (f_to_g prw)"
  let ?cpx = "redtriple.cpx ?rp"
  let ?all = "s_list @ ns_list @ nst_list"
  let ?ws = "redtriple.not_ws_ns ?rp"
  obtain ch "pr" w w0 least scf where id: "?pr = (ch,pr,w,w0,least,scf)" by (cases ?pr, force)
  assume valid: "isOK(redtriple.valid ?rp)"
   and stri: "isOK(check_allm (redtriple.s ?rp) s_list)"
   and nstri: "isOK (check_allm (redtriple.ns ?rp) ns_list)"
   and nstri_top: "isOK (check_allm (redtriple.nst ?rp) nst_list)"
  have af: "?af = full_af" "?af' = full_af" by (simp_all add: id Let_def)
  from valid have "isOK(ch)" by (simp add: id Let_def)
  hence ch: "ch = succeed" by (cases ch, auto)
  have cpx: "?cpx = no_complexity" by (simp add: id Let_def)
  note id = id[unfolded ch cpx]
  interpret admissible_weight_fun_prc w w0 "(\<lambda> fn gm. pr fn > pr gm)" "(\<lambda> fn gm. pr fn \<ge> pr gm)" least scf
    by (rule prec_weight_repr_to_prec_weight[OF id])
  from kbo_redpair 
  interpret mono_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af .
  show "\<exists> S NS NST. cpx_ce_af_redtriple_order S NS NST ?af ?af' ?cpx 
    \<and> set s_list \<subseteq> S \<and> set ns_list \<subseteq> NS \<and> set nst_list \<subseteq> NST 
    \<and> not_ws_info NS (redtriple.not_ws_ns ?rp)
    \<and> (isOK (redtriple.mono ?rp ?all) \<longrightarrow> mono_ce_af_redtriple_order S NS NST ?af \<and> ctxt_closed S)"
    unfolding af cpx
  proof (intro exI conjI impI)
    show "cpx_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af full_af no_complexity"
      by (unfold_locales, rule ctxt_closed_imp_af_monotone[OF kbo_S_mono])
    show "mono_ce_af_redtriple_order kbo_S kbo_NS kbo_NS full_af" ..
    show "ctxt_closed kbo_S" by (rule kbo_S_mono)
    show "set s_list \<subseteq> kbo_S" using stri by (force simp: Let_def id)
    show "set ns_list \<subseteq> kbo_NS" using nstri by (force simp: Let_def id)
    show "set nst_list \<subseteq> kbo_NS" using nstri_top by (force simp: Let_def id)
    show "not_ws_info kbo_NS ?ws" 
      by (simp add: id Let_def, intro allI ws_fun_argI kbo_NS_supteq, auto)
  qed
qed

end
