(*
Author:  Christian Sternagel <c.sternagel@gmail.com> (2013-2015)
Author:  Martin Avanzini <martin.avanzini@uibk.ac.at> (2014)
Author:  René Thiemann <rene.thiemann@uibk.ac.at> (2013-2015)
License: LGPL (see file COPYING.LESSER)
*)

chapter \<open>Multihole Contexts\<close>

theory Multihole_Context
imports 
  Trs 
  "../Auxiliaries/Util"
  "../Auxiliaries/SubList"
  "~~/src/HOL/Library/Lattice_Syntax"
begin

(*TODO: move*)
lemma listsum_take [simp]:
  "i < length (xs :: nat list) \<Longrightarrow> listsum (take i xs) \<le> listsum xs"
  by (induct xs arbitrary: i) (simp, case_tac i, auto)

(*TODO: move*)
lemma concat_nth:
  assumes "m < length xs" and "n < length (xs ! m)"
    and "i = listsum (map length (take m xs)) + n"
  shows "concat xs ! i = xs ! m ! n"
using assms
proof (induct xs arbitrary: m n i)
  case (Cons x xs)
  show ?case
  proof (cases m)
    case 0
    then show ?thesis using Cons by (simp add: nth_append)
  next
    case (Suc k)
    with Cons(1) [of k n "i - length x"] and Cons(2-)
      show ?thesis by (simp_all add: nth_append)
  qed
qed simp

subsection \<open>Partitioning lists into chunks of given length\<close>

fun partition_by
where
  "partition_by xs [] = []" |
  "partition_by xs (y#ys) = take y xs # partition_by (drop y xs) ys"

lemma concat_partition_by [simp]:
  "listsum ys = length xs \<Longrightarrow> concat (partition_by xs ys) = xs"
  by (induct ys arbitrary: xs) simp_all

lemma partition_by_nth_nth:
  assumes "i < length (partition_by xs ys)"
    and "j < length (partition_by xs ys ! i)"
    and "listsum ys = length xs"
  shows "partition_by xs ys ! i ! j = xs ! (listsum (map length (take i (partition_by xs ys))) + j)"
  using concat_nth [OF assms(1, 2) refl]
  unfolding concat_partition_by [OF assms(3)] by simp

lemma length_partition_by [simp]:
  "length (partition_by xs ys) = length ys"
  by (induct ys arbitrary: xs) simp_all

lemma partition_by_Nil [simp]:
  "partition_by [] ys = replicate (length ys) []"
  by (induct ys) simp_all

lemma partition_by_concat_id [simp]:
  assumes "length xss = length ys"
    and "\<And>i. i < length ys \<Longrightarrow> length (xss ! i) = ys ! i"
  shows "partition_by (concat xss) ys = xss"
  using assms by (induct ys arbitrary: xss) (simp, case_tac xss, simp, fastforce)

lemma partition_by_nth:
  "i < length ys \<Longrightarrow> partition_by xs ys ! i = take (ys ! i) (drop (listsum (take i ys)) xs)"
  by (induct ys arbitrary: xs i) (simp, case_tac i, simp_all add: ac_simps)

lemma length_partition_by_nth [simp]:
  "listsum ys = length xs \<Longrightarrow> i < length ys \<Longrightarrow> length (partition_by xs ys ! i) = ys ! i"
  by (induct ys arbitrary: xs i) (simp, case_tac i, simp_all)

lemma map_length_partition_by [simp]:
  "listsum ys = length xs \<Longrightarrow> map length (partition_by xs ys) = ys"
by (simp add: map_nth_eq_conv)

lemma map_partition_by_nth [simp]:
  "i < length ys \<Longrightarrow> map f (partition_by xs ys ! i) = partition_by (map f xs) ys ! i"
  by (induct ys arbitrary: i xs) (simp, case_tac i, simp_all add: take_map drop_map)

lemma listsum_partition_by [simp]:
  "listsum ys = length xs \<Longrightarrow>
    listsum (map (\<lambda>x. listsum (map f x)) (partition_by xs ys)) = listsum (map f xs)"
  by (induct ys arbitrary: xs) (simp_all, metis append_take_drop_id listsum_append map_append)

lemma partition_by_map_conv:
  "partition_by xs ys = map (\<lambda>i. take (ys ! i) (drop (listsum (take i ys)) xs)) [0 ..< length ys]"
  by (rule nth_equalityI) (simp_all add: partition_by_nth)

lemma UN_set_partition_by_map:
  "listsum ys = length xs \<Longrightarrow> (\<Union>x\<in>set (partition_by (map f xs) ys). \<Union>set x) = \<Union>set (map f xs)"
  by (induct ys arbitrary: xs)
     (simp_all add: drop_map take_map, metis UN_Un append_take_drop_id set_append)

lemma UN_set_partition_by:
  "listsum ys = length xs \<Longrightarrow> (\<Union>zs \<in> set (partition_by xs ys). \<Union>x \<in> set zs. f x) = (\<Union>x \<in> set xs. f x)"
  by (induct ys arbitrary: xs) (simp_all, metis UN_Un append_take_drop_id set_append)

lemma Ball_atLeast0LessThan_partition_by_conv:
  "(\<forall>i\<in>{0..<length ys}. \<forall>x\<in>set (partition_by xs ys ! i). P x) =
    (\<forall>x \<in> \<Union>(set (map set (partition_by xs ys))). P x)"
  by auto (metis atLeast0LessThan in_set_conv_nth length_partition_by lessThan_iff)

lemma Ball_set_partition_by:
  "listsum ys = length xs \<Longrightarrow>
  (\<forall>x \<in> set (partition_by xs ys). \<forall>y \<in> set x. P y) = (\<forall>x \<in> set xs. P x)"
proof (induct ys arbitrary: xs)
  case (Cons y ys)
  then show ?case
    apply (subst (2) append_take_drop_id [of y xs, symmetric])
    apply (simp only: set_append)
    apply auto
  done
qed simp

datatype ('f, vars_mctxt : 'v) mctxt = MVar 'v | MHole | MFun 'f "('f, 'v) mctxt list"


subsection \<open>Conversions from and to multihole contexts\<close>

primrec mctxt_of_term :: "('f, 'v) term \<Rightarrow> ('f, 'v) mctxt"
where
  "mctxt_of_term (Var x) = MVar x" |
  "mctxt_of_term (Fun f ts) = MFun f (map mctxt_of_term ts)"

primrec term_of_mctxt :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v) term"
where
  "term_of_mctxt (MVar x) = Var x" |
  "term_of_mctxt (MFun f Cs) = Fun f (map term_of_mctxt Cs)"

lemma term_of_mctxt_mctxt_of_term_id [simp]:
  "term_of_mctxt (mctxt_of_term t) = t"
  by (induct t) (simp_all add: map_idI)

fun num_holes :: "('f, 'v) mctxt \<Rightarrow> nat"
where
  "num_holes (MVar _) = 0" |
  "num_holes MHole = 1" |
  "num_holes (MFun _ ctxts) = listsum (map num_holes ctxts)"

lemma mctxt_of_term_term_of_mctxt_id [simp]:
  "num_holes C = 0 \<Longrightarrow> mctxt_of_term (term_of_mctxt C) = C"
  by (induct C) (simp_all add: map_idI)

lemma num_holes_mctxt_of_term [simp]:
  "num_holes (mctxt_of_term t) = 0"
  by (induct t) simp_all

fun funas_mctxt :: "('f, 'v) mctxt \<Rightarrow> 'f sig"
where
  "funas_mctxt (MFun f Cs) = {(f, length Cs)} \<union> \<Union>(funas_mctxt ` set Cs)" |
  "funas_mctxt _ = {}"

fun funas_mctxt_list :: "('f, 'v) mctxt \<Rightarrow> ('f \<times> nat) list"
where
  "funas_mctxt_list (MFun f Cs) = (f, length Cs) # concat (map funas_mctxt_list Cs)" |
  "funas_mctxt_list _ = []"

lemma funas_mctxt_list [simp]:
  "set (funas_mctxt_list C) = funas_mctxt C"
  by (induct C) simp_all

fun split_term :: "(('f, 'v) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) mctxt \<times> ('f, 'v) term list"
where
  "split_term P (Var x) = (if P (Var x) then (MHole, [Var x]) else (MVar x, []))" |
  "split_term P (Fun f ts) =
    (if P (Fun f ts) then (MHole, [Fun f ts])
    else let us = map (split_term P) ts in (MFun f (map fst us), concat (map snd us)))"

fun cap_till :: "(('f, 'v) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) mctxt"
where
  "cap_till P (Var x) = (if P (Var x) then MHole else MVar x)" |
  "cap_till P (Fun f ts) = (if P (Fun f ts) then MHole else MFun f (map (cap_till P) ts))"

fun uncap_till :: "(('f, 'v) term \<Rightarrow> bool) \<Rightarrow> ('f, 'v) term \<Rightarrow> ('f, 'v) term list"
where
  "uncap_till P (Var x) = (if P (Var x) then [Var x] else [])" |
  "uncap_till P (Fun f ts) = (if P (Fun f ts) then [Fun f ts] else concat (map (uncap_till P) ts))"

lemma split_term [simp]:
  "split_term P t = (cap_till P t, uncap_till P t)"
  by (induct t) (simp_all cong: map_cong)

definition "if_Fun_in_set F = (\<lambda>t. is_Var t \<or> the (root t) \<in> F)"

lemma if_Fun_in_set_simps [simp]:
  "if_Fun_in_set F (Var x)"
  "if_Fun_in_set F (Fun f ts) \<longleftrightarrow> (f, length ts) \<in> F"
  "is_Var t \<Longrightarrow> if_Fun_in_set F t"
  "is_Fun t \<Longrightarrow> if_Fun_in_set F t \<longleftrightarrow> the (root t) \<in> F"
  by (simp_all add: if_Fun_in_set_def)

lemma if_Fun_in_set_mono:
  "F \<subseteq> G \<Longrightarrow> if_Fun_in_set F t \<Longrightarrow> if_Fun_in_set G t"
  by (auto simp: if_Fun_in_set_def)

abbreviation "split_term_funas F \<equiv> split_term (if_Fun_in_set F)"
abbreviation "cap_till_funas F \<equiv> cap_till (if_Fun_in_set F)"
abbreviation "uncap_till_funas F \<equiv> uncap_till (if_Fun_in_set F)"

lemma if_Fun_in_set_uncap_till_funas:
  "A \<subseteq> B \<Longrightarrow> if_Fun_in_set A t \<Longrightarrow> uncap_till_funas B t = [t]"
  by (cases t) auto

lemma cap_till_funasD [dest]:
  "fn \<in> funas_mctxt (cap_till_funas F t) \<Longrightarrow> fn \<in> F \<Longrightarrow> False"
proof (induct t)
  case (Fun f ts)
  then show ?case by (cases "(f, length ts) \<in> F") auto
qed simp

lemma cap_till_funas:
  "\<forall>fn \<in> funas_mctxt (cap_till_funas F t). fn \<notin> F"
  by auto

lemma uncap_till:
  "\<forall>s \<in> set (uncap_till P t). P s"
  by (induct t) simp_all

(*do not add to simpset, since it severely degrades simplifier performance*)
lemma uncap_till_singleton:
  assumes "s \<in> set (uncap_till P t)"
  shows "uncap_till P s = [s]"
using assms
proof (induct t)
  case (Fun f ts)
  then show ?case by (cases "P (Fun f ts)") auto
qed simp

lemma uncap_till_idemp [simp]:
  "map (uncap_till P) (uncap_till P t) = map (\<lambda>s. [s]) (uncap_till P t)"
  using assms by (intro map_cong [OF refl] uncap_till_singleton) simp_all

lemma uncap_till_Fun [simp]:
  "P (Fun f ts) \<Longrightarrow> uncap_till P (Fun f ts) = [Fun f ts]"
  by simp

fun fill_holes :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) term"
where
  "fill_holes (MVar x) _ = Var x" |
  "fill_holes MHole [t] = t" |
  "fill_holes (MFun f cs) ts = (Fun f (map (\<lambda> i. fill_holes (cs ! i)
    (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts))) [0 ..< length cs]))"

lemma fill_holes_MHole:
  "length ts = 1 \<Longrightarrow> ts ! 0 = u \<Longrightarrow> fill_holes MHole ts = u"
  by (cases ts) simp_all

abbreviation "partition_holes xs Cs \<equiv> partition_by xs (map num_holes Cs)"

(*some compatibility lemmas (which should be dropped eventually)*)
lemmas
  map_partition_holes_nth [simp] =
    map_partition_by_nth [of _ "map num_holes Cs" for Cs, unfolded length_map] and
  length_partition_holes [simp] =
    length_partition_by [of _ "map num_holes Cs" for Cs, unfolded length_map]

lemma partition_holes_nth:
  "i < length Cs \<Longrightarrow>
    partition_holes xs Cs ! i =
      take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) xs)"
  using partition_by_nth [of i "map num_holes Cs" xs] by (simp add: take_map)

lemma length_partition_holes_nth [simp]:
  assumes "listsum (map num_holes cs) = length ts"
    and "i < length cs"
  shows "length (partition_holes ts cs ! i) = num_holes (cs ! i)"
  using assms by simp

lemma partition_holes_def:
  "partition_holes ts cs = map (\<lambda>i.
    take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts)) [0 ..< length cs]"
  by (simp add: partition_by_map_conv [of ts "map num_holes cs"] take_map)

lemma concat_partition_holes_upt:
  assumes "i \<le> length cs"
  shows "concat [partition_holes ts cs ! j. j \<leftarrow> [0 ..< i]] =
    take (listsum [num_holes (cs ! j). j \<leftarrow> [0 ..< i]]) ts"
using assms
proof (induct i arbitrary: ts)
  case (Suc i)
  hence i': "i < length cs" by (metis less_eq_Suc_le)
  then have *: "i < length (map num_holes cs)" by simp
  hence i'':  "i \<le> length cs" by auto
  show ?case
    unfolding upt_Suc_append[OF le0] map_append concat_append Suc(1)[OF i''] concat.simps append_Nil2
    unfolding listsum_append take_add 
    unfolding list.map(2)
    unfolding partition_by_nth [OF *]
    unfolding take_map nth_map [OF i']
    unfolding take_upt_idx[OF i']
    unfolding map_map o_def by auto
qed (auto)


lemma partition_holes_step:
  "partition_holes ts (C # Cs) = take (num_holes C) ts # partition_holes (drop (num_holes C) ts) Cs"
  by simp

lemma partition_holes_map_ctxt:
  assumes "length cs = length ds"
    and "\<And> i. i < length cs \<Longrightarrow> num_holes (cs ! i) = num_holes (ds ! i)"
  shows "partition_holes ts cs = partition_holes ts ds"
  using assms by (metis nth_map_conv)

lemma partition_holes_concat_id:
  assumes "length sss = length cs" 
    and "\<And> i. i < length cs \<Longrightarrow> num_holes (cs ! i) = length (sss ! i)"
  shows "partition_holes (concat sss) cs = sss"
  using assms by (intro partition_by_concat_id) auto

lemma partition_holes_fill_holes_conv:
  "fill_holes (MFun f cs) ts =
    Fun f [fill_holes (cs ! i) (partition_holes ts cs ! i). i \<leftarrow> [0 ..< length cs]]"
  by (simp add: partition_by_nth take_map)

lemma fill_holes_arbitrary:
  assumes lCs: "length Cs = length ts"
    and lss: "length ss = length ts"
    and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> f (Cs ! i) (ss ! i) = ts ! i"
  shows "map (\<lambda>i. f (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) (concat ss)))) [0 ..< length Cs]
    = ts"
proof -
  have l: "length [0 ..< length Cs] = length ts" unfolding lCs by simp
  show ?thesis 
  proof (unfold map_nth_eq_conv[OF l], intro allI impI)
    fix i
    assume i: "i < length ts"
    with lCs have iC: "i < length Cs" by simp
    hence "f (Cs ! ([0..<length Cs] ! i))
         (take (num_holes (Cs ! ([0..<length Cs] ! i))) (drop (listsum (map num_holes (take ([0..<length Cs] ! i) Cs))) (concat ss)))
       = f (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) (concat ss)))" (is "?l = _")
     by simp
    also have "\<dots> = ts ! i" using i lCs lss rec
    proof (induct i arbitrary: ts Cs ss)
      case 0
      from 0 obtain t ts' where ts: "ts = t # ts'" by (cases ts, auto)
      from 0 obtain s ss' where ss: "ss = s # ss'" by (cases ss, auto)
      from 0 obtain C Cs' where Cs: "Cs = C # Cs'" by (cases Cs, auto)
      show ?case using 0(4)[of 0] unfolding ts ss Cs by auto
    next
      case (Suc i)
      from Suc obtain t ts' where ts: "ts = t # ts'" by (cases ts, auto)
      from Suc obtain s ss' where ss: "ss = s # ss'" by (cases ss, auto)
      from Suc obtain C Cs' where Cs: "Cs = C # Cs'" by (cases Cs, auto)
      from Suc ts ss Cs have i: "i < length ts'" "length Cs' = length ts'" "length ss' = length ts'" by auto
      from Suc(5)[of 0] ts ss Cs have nh: "num_holes C = length s" "f C s = t" by auto
      show ?case using Suc(5) unfolding Cs ss ts using nh
        by (force intro!: Suc(1) i)
    qed
    finally show "?l = ts ! i" .
  qed
qed

lemma fill_holes_MFun:
  assumes lCs: "length Cs = length ts"
    and lss: "length ss = length ts"
    and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> fill_holes (Cs ! i) (ss ! i) = ts ! i"
  shows "fill_holes (MFun f Cs) (concat ss) = Fun f ts" 
  unfolding fill_holes.simps term.simps
    by (rule conjI[OF refl], rule fill_holes_arbitrary[OF lCs lss rec])

inductive
  eq_fill ::
    "('f, 'v) term \<Rightarrow> ('f, 'v) mctxt \<times> ('f, 'v) term list \<Rightarrow> bool" ("(_/ =\<^sub>f _)" [51, 51] 50)
where
  eqfI [intro]: "t = fill_holes D ss \<Longrightarrow> num_holes D = length ss \<Longrightarrow> t =\<^sub>f (D, ss)"

lemma fill_holes_inj:
  assumes "num_holes C = length ss"
    and "length ts = length ss"
    and "fill_holes C ss = fill_holes C ts"
  shows "ss = ts"
using assms
proof (induction C arbitrary: ss ts)
  case (MFun f Cs)
  then have [simp]: "listsum (map num_holes Cs) = length ss" "length ts = length ss" by auto
  let ?ss = "partition_holes ss Cs"
  let ?ts = "partition_holes ts Cs"
  { fix i
    assume "i < length Cs"
    assume "num_holes (Cs ! i) = length (?ss ! i)"
      and "length (?ts ! i) = length (?ss ! i)"
      and "fill_holes (Cs ! i) (?ss ! i) = fill_holes (Cs ! i) (?ts ! i)"
    moreover then have "Cs ! i \<in> set Cs" using \<open>i < length Cs\<close> by auto
    ultimately have "?ss ! i = ?ts ! i" using MFun.IH by auto }
  note IH = this
  from MFun.prems have "\<forall>i<length Cs. fill_holes (Cs ! i) (?ss ! i) = fill_holes (Cs ! i) (?ts ! i)"
    and "\<forall>i<length Cs. num_holes (Cs ! i) = length (?ss ! i)"
    and "\<forall>i<length Cs. length (?ts ! i) = length (?ss ! i)"
    unfolding partition_holes_fill_holes_conv by auto
  with IH have "\<forall>i<length Cs. ?ss ! i = ?ts ! i" by (blast)
  then have "?ss = ?ts" by (intro nth_equalityI) (simp_all)
  then have "concat ?ss = concat ?ts" by auto
  then show ?case by (simp)
next
  case (MVar x)
  then show ?case by (auto elim!: eq_fill.cases)
next
  case (MHole)
  then show ?case
  by (auto elim!: eq_fill.cases simp: fill_holes_MHole) (metis One_nat_def less_one nth_equalityI)
qed

lemma eqf_refl [intro]:
  "num_holes C = length ts \<Longrightarrow> fill_holes C ts =\<^sub>f (C, ts)"
  by (auto)

lemma eqfE:
  assumes "t =\<^sub>f (D, ss)" shows "t = fill_holes D ss" "num_holes D = length ss"
  using assms[unfolded eq_fill.simps] by auto

lemma eqf_MFunI:
  assumes "length sss = length Cs"
    and "length ts = length Cs"
    and"\<And> i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)"
  shows "Fun f ts =\<^sub>f (MFun f Cs, concat sss)"
proof 
  have "num_holes (MFun f Cs) = listsum (map num_holes Cs)" by simp
  also have "map num_holes Cs = map length sss"
    by (rule nth_equalityI, insert assms eqfE[OF assms(3)], auto)
  also have "listsum (\<dots>) = length (concat sss)" unfolding length_concat ..
  finally show "num_holes (MFun f Cs) = length (concat sss)" .
  show "Fun f ts = fill_holes (MFun f Cs) (concat sss)"
    by (rule fill_holes_MFun[symmetric], insert assms(1,2) eqfE[OF assms(3)], auto)
qed

lemma eqf_MFunE:
  assumes "s =\<^sub>f (MFun f Cs,ss)"  
  obtains ts sss where "s = Fun f ts" "length ts = length Cs" "length sss = length Cs" 
  "\<And> i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)"
  "ss = concat sss" 
proof -
  from eqfE[OF assms] have fh: "s = fill_holes (MFun f Cs) ss" 
    and nh: "listsum (map num_holes Cs) = length ss" by auto 
  from fh obtain ts where s: "s = Fun f ts" by (cases s, auto)
  from fh[unfolded s] 
  have ts: "ts = map (\<lambda>i. fill_holes (Cs ! i) (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ss))) [0..<length Cs]" 
    (is "_ = map (?f Cs ss) _")
    by auto
  from nh 
  have "\<exists> sss. length sss = length Cs \<and> (\<forall> i < length Cs . ts ! i =\<^sub>f (Cs ! i, sss ! i)) \<and> ss = concat sss"
    unfolding ts
  proof (induct Cs arbitrary: ss)
    case (Cons C Cs ss)
    let ?n = "num_holes C"
    let ?bef = "take ?n ss"
    let ?aft = "drop ?n ss"
    def bef \<equiv> ?bef
    def aft \<equiv> ?aft
    from Cons(2) have "length ?bef = ?n" "listsum (map num_holes Cs) = length ?aft" by auto
    hence lb: "?n = length bef" "listsum (map num_holes Cs) = length aft" unfolding bef_def aft_def by auto
    have ss: "ss = bef @ aft" unfolding bef_def aft_def by auto
    have l: "length (C # Cs) = Suc (length Cs)" by simp
    note IH = Cons(1)[OF lb(2)]
    from IH obtain sss where len: "length sss = length Cs" and
      i: "\<And> i. i < length Cs \<Longrightarrow> map (?f Cs aft) [0 ..< length Cs] ! i =\<^sub>f (Cs ! i, sss ! i)" and
      aft: "aft = concat sss" by blast
    have "?case \<longleftrightarrow> (\<exists>sss. length sss = Suc (length Cs) \<and>
          fill_holes C bef =\<^sub>f (C, sss ! 0) \<and>
          (\<forall>i<length Cs.
              ?f Cs aft i =\<^sub>f (Cs ! i, sss ! Suc i)) \<and>
          bef @ aft = concat sss)"
      unfolding l map_upt_Suc all_Suc_conv 
      unfolding ss by (simp add: lb)
    also have \<dots>
    proof (rule exI[of _ "bef # sss"], intro conjI allI impI)
      show "length (bef # sss) = Suc (length Cs)" by (simp add: len)
      show "fill_holes C bef =\<^sub>f (C, (bef # sss) ! 0)" using eqfI[OF _ lb(1)] by simp
      show "bef @ aft = concat (bef # sss)" using aft by simp
      fix i
      assume "i < length Cs"
      with i[OF this]
      show "?f Cs aft i =\<^sub>f (Cs ! i, (bef # sss) ! Suc i)" by simp
    qed        
    finally show ?case by simp
  qed auto
  then obtain sss where *: "length sss = length Cs" "\<And> i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)"
    "ss = concat sss" by blast
  have len: "length ts = length Cs" unfolding ts by auto
  assume ass: "\<And>ts sss. s = Fun f ts \<Longrightarrow>
              length ts = length Cs \<Longrightarrow>
              length sss = length Cs \<Longrightarrow> (\<And>i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)) \<Longrightarrow> ss = concat sss \<Longrightarrow> thesis"
  show thesis
    by (rule ass[OF s len *])
qed

fun mctxt_of_ctxt :: "('f, 'v) ctxt \<Rightarrow> ('f, 'v) mctxt"
where
  "mctxt_of_ctxt Hole = MHole" |
  "mctxt_of_ctxt (More f ss\<^sub>1 C ss\<^sub>2) =
    MFun f (map mctxt_of_term ss\<^sub>1 @ mctxt_of_ctxt C # map mctxt_of_term ss\<^sub>2)"

lemma num_holes_mctxt_of_ctxt [simp]:
  "num_holes (mctxt_of_ctxt C) = 1"
  by (induct C) simp_all

lemma mctxt_of_term: "t =\<^sub>f (mctxt_of_term t, [])"
proof (induct t)
  case (Var x)
  show ?case by (rule, auto)
next
  case (Fun f ts)
  let ?ss = "map (\<lambda> _. []) ts"
  have id: "concat ?ss = []" by simp
  have "?case = (Fun f ts =\<^sub>f (MFun f (map mctxt_of_term ts), concat ?ss))" unfolding id by simp
  also have "\<dots>"
    by (rule eqf_MFunI, insert Fun[unfolded set_conv_nth], auto)
  finally show ?case .
qed

lemma mctxt_of_ctxt [simp]:
  "C\<langle>t\<rangle> =\<^sub>f (mctxt_of_ctxt C, [t])"
proof (induct C)
  case Hole
  show ?case by (rule, auto)
next
  case (More f bef C aft)
  let ?sss = "map (\<lambda> _. []) bef @ [t] # map (\<lambda> _. []) aft"
  let ?ts = "map mctxt_of_term bef @ mctxt_of_ctxt C # map mctxt_of_term aft"
  have id: "concat ?sss = [t]" by (induct bef, auto)
  have "?case = 
    (Fun f (bef @ C\<langle>t\<rangle> # aft) =\<^sub>f (MFun f ?ts, concat ?sss))"
    unfolding id by simp
  also have "\<dots>"
  proof (rule eqf_MFunI)
    fix i
    assume i: "i < length ?ts"
    show "(bef @ C\<langle>t\<rangle> # aft) ! i =\<^sub>f (?ts ! i, ?sss ! i)"
      using More i
      by (cases "i < length bef", simp add: nth_append mctxt_of_term,
        cases "i = length bef", auto simp: nth_append mctxt_of_term)
  qed auto
  finally show ?case .
qed

lemma fill_holes_ctxt_main:
  assumes "num_holes C = Suc (length bef + length aft)"
  shows "\<exists> D. \<forall> s. fill_holes C (bef @ s # aft) = D \<langle> s \<rangle>"
using assms
proof (induct C arbitrary: bef aft)
  case MHole
  show ?case
    by (rule exI[of _ \<box>], insert MHole, auto)
next
  case (MFun f cs)
  note IH = MFun(1)
  note holes = MFun(2)
  let ?p = "\<lambda> bef aft b a D cs s. map (\<lambda>i. fill_holes (cs ! i)
                        (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) (bef @ s # aft))))
               [0..<length cs] =
              b @ D\<langle>s\<rangle> # a"
  from holes IH
  have "\<exists> b D a. \<forall>s. ?p bef aft b a D cs s"
  proof (induct cs arbitrary: bef)
    case (Cons c ccs)
    have len: "length (c # ccs) = Suc (length ccs)" by simp
    show ?case 
    proof (cases "num_holes c \<le> length bef")
      case True
      hence "bef = take (num_holes c) bef @ drop (num_holes c) bef 
        \<and> length (take (num_holes c) bef) = num_holes c" by auto
      then obtain bc ba where bef: "bef = bc @ ba" and lbc: "length bc = num_holes c" by blast
      from Cons(2) have nh: "num_holes (MFun f ccs) = Suc (length ba + length aft)" unfolding bef
        by (simp add: lbc)
      from Cons(1)[OF nh Cons(3)] obtain b D a where IH: "\<And> s. ?p ba aft b a D ccs s" by auto
      show ?thesis unfolding len map_upt_Suc bef
        by (rule exI[of _ "fill_holes c bc # b"], rule exI[of _ D], rule exI[of _ a], insert IH lbc, auto)
    next
      case False
      hence "\<exists> la. num_holes c = Suc (length bef + la)" by arith
      then obtain la where nhc: "num_holes c = Suc (length bef + la)" ..
      from Cons(2) nhc have "length (take la aft) = la" by auto
      from Cons(3)[of c bef "take la aft", unfolded this, OF _ nhc]
      obtain D where D: "\<forall>s. fill_holes c (bef @ s # take la aft) = D\<langle>s\<rangle>" by auto
      show ?thesis unfolding len map_upt_Suc
        by (rule exI[of _ Nil], rule exI[of _ D], simp add: nhc D) 
    qed
  qed auto
  then obtain b D a where main: "\<And> s. ?p bef aft b a D cs s" by blast
  show ?case by (rule exI[of _ "More f b D a"], insert main, auto)
qed simp

lemma fill_holes_ctxt:
  assumes nh: "num_holes C = length ss"
    and i: "i < length ss"
  obtains D where "\<And> s. fill_holes C (ss[i := s]) = D \<langle> s \<rangle>"
proof -
  from id_take_nth_drop[OF i] obtain bef aft where ss: "ss = bef @ ss ! i # aft" 
    and bef: "bef = take i ss" by blast
  from bef i have bef: "length bef = i" by auto
  note len = arg_cong[OF ss, of length]
  from len nh
  have "num_holes C = Suc (length bef + length aft)" by simp
  from fill_holes_ctxt_main[OF this] obtain D where id: "\<And> s. fill_holes C (bef @ s # aft) = D \<langle> s \<rangle>" by blast
  {
    fix s
    have "ss[i := s] = bef @ s # aft" unfolding arg_cong[OF ss, of "\<lambda> ss. ss[i := s]"]
      using i bef by auto
    with id[of s] have "fill_holes C (ss[i := s]) = D \<langle>s \<rangle>" by simp
  }
  hence main: "\<exists> D. \<forall> s. fill_holes C (ss[i := s]) = D \<langle>s \<rangle>" by blast
  assume "\<And> D. \<lbrakk>\<And>s. fill_holes C (ss[i := s]) = D\<langle>s\<rangle>\<rbrakk> \<Longrightarrow> thesis"
  with main show thesis by blast
qed

fun map_vars_term_mctxt :: "('v \<Rightarrow> 'w) \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> ('f, 'w) mctxt"
where
  "map_vars_term_mctxt vw MHole = MHole" |
  "map_vars_term_mctxt vw (MVar v) = (MVar (vw v))" |
  "map_vars_term_mctxt vw (MFun f Cs) = MFun f (map (map_vars_term_mctxt vw) Cs)"

lemma map_vars_term_mctxt_id [simp]:
  "map_vars_term_mctxt (\<lambda> x. x) C = C"
  by (induct C, auto intro: nth_equalityI)

lemma num_holes_map_vars_term_mctxt [simp]:
  "num_holes (map_vars_term_mctxt vw C) = num_holes C"
proof (induct C)
  case (MFun f Cs)
  thus ?case by (induct Cs, auto)
qed auto

lemma map_vars_term_eq_fill:
  "t =\<^sub>f (C,ss) \<Longrightarrow> map_vars_term vw t =\<^sub>f (map_vars_term_mctxt vw C, map (map_vars_term vw) ss)"
proof (induct C arbitrary: t ss)
  case (MFun f Cs s ss)
  from eqf_MFunE[OF MFun(2)] obtain ts sss where s: "s = Fun f ts" and len: "length ts = length Cs" "length sss = length Cs" 
    and IH: "\<And> i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)" and ss: "ss = concat sss" by metis
  {
    fix i
    assume i: "i < length Cs"
    hence "Cs ! i \<in> set Cs" by auto
    from MFun(1)[OF this IH[OF i]] have "map_vars_term vw (ts ! i) =\<^sub>f (map_vars_term_mctxt vw (Cs ! i), map (map_vars_term vw) (sss ! i))" .
  } note IH = this
  show ?case unfolding map_vars_term_mctxt.simps ss map_concat s term.map
    by (rule eqf_MFunI, insert IH len, auto)
next
  case (MHole t ss)
  from eqfE[OF this]
  show ?case by (cases ss, auto)
next
  case (MVar v t ss)
  from eqfE[OF this]
  show ?case by (cases ss, auto)
qed

lemma map_vars_term_fill_holes:
  assumes nh: "num_holes C = length ss"
  shows "map_vars_term vw (fill_holes C ss) =
    fill_holes (map_vars_term_mctxt vw C) (map (map_vars_term vw) ss)"
proof -
  from eqfE[OF map_vars_term_eq_fill[OF eqfI[OF refl nh]]]
  show ?thesis by simp
qed

lemma split_term_eqf:
  "t =\<^sub>f (cap_till P t, uncap_till P t)"
proof (induct t)
  case (Fun f ts)
  show ?case
  proof (cases "P (Fun f ts)")
    case False
    then have "?thesis = (Fun f ts =\<^sub>f (MFun f (map (cap_till P) ts), concat (map (uncap_till P) ts)))"
      by simp
    also have "\<dots>"
    proof (rule eqf_MFunI)
      fix i
      presume "i < length ts"
      moreover then have "ts ! i \<in> set ts" by auto
      ultimately show "ts ! i =\<^sub>f (map (cap_till P) ts ! i, map (uncap_till P) ts ! i)"
        using Fun by auto
    qed simp_all
    finally show ?thesis .
  qed auto
qed auto

lemma fill_holes_cap_till_uncap_till_id [simp]:
  "fill_holes (cap_till P t) (uncap_till P t) = t"
proof -
  have "t =\<^sub>f (cap_till P t, uncap_till P t)" by (metis split_term_eqf)
  from eqfE [OF this] show ?thesis by simp
qed

lemma num_holes_cap_till [simp]:
  "num_holes (cap_till P t) = length (uncap_till P t)"
  using eqfE [OF split_term_eqf] by auto

fun split_vars :: "('f, 'v) term \<Rightarrow> (('f, 'v) mctxt \<times> 'v list)"
where
  "split_vars (Var x) = (MHole, [x])" |
  "split_vars (Fun f ts) = (MFun f (map (fst \<circ> split_vars) ts), concat (map (snd \<circ> split_vars) ts))"

lemma split_vars_num_holes: "num_holes (fst (split_vars t)) = length (snd (split_vars t))"
proof (induct t)
  case (Fun f ts)
  thus ?case by (induct ts, auto)
qed simp

lemma split_vars_vars_term_list: "snd (split_vars t) = vars_term_list t"
proof (induct t)
  case (Fun f ts)
  thus ?case by (auto simp: vars_term_list.simps o_def, induct ts, auto)
qed (auto simp: vars_term_list.simps)

lemma split_vars_vars_term: "set (snd (split_vars t)) = vars_term t"
  using arg_cong[OF split_vars_vars_term_list[of t], of set] by auto

lemma split_vars_eqf_subst_map_vars_term:
  "t \<cdot> \<sigma> =\<^sub>f (map_vars_term_mctxt vw (fst (split_vars t)), map \<sigma> (snd (split_vars t)))"
proof (induct t)
  case (Fun f ts)
  have "?case = (Fun f (map (\<lambda>t. t \<cdot> \<sigma>) ts)
    =\<^sub>f (MFun f (map (map_vars_term_mctxt vw \<circ> (fst \<circ> split_vars)) ts), concat (map (map \<sigma> \<circ> (snd \<circ> split_vars)) ts)))"
    by (simp add: map_concat)
  also have "..." 
  proof (rule eqf_MFunI, simp, simp, unfold length_map)
    fix i
    assume i: "i < length ts"
    hence mem: "ts ! i \<in> set ts" by auto
    show "map (\<lambda>t. t \<cdot> \<sigma>) ts ! i =\<^sub>f (map (map_vars_term_mctxt vw \<circ> (fst \<circ> split_vars)) ts ! i, map (map \<sigma> \<circ> (snd \<circ> split_vars)) ts ! i)"
      using Fun[OF mem] i by auto
  qed
  finally show ?case by simp
qed auto

lemma split_vars_eqf_subst: "t \<cdot> \<sigma> =\<^sub>f (fst (split_vars t), (map \<sigma> (snd (split_vars t))))"
  using split_vars_eqf_subst_map_vars_term[of t \<sigma> "\<lambda> x. x"] by simp

lemma split_vars_into_subst_map_vars_term:
  assumes split: "split_vars l = (C,xs)"
    and len: "length ts = length xs"
    and id: "\<And> i. i < length xs \<Longrightarrow> \<sigma> (xs ! i) = ts ! i"
  shows "l \<cdot> \<sigma> =\<^sub>f (map_vars_term_mctxt vw C,ts)"
proof -
  from split_vars_eqf_subst_map_vars_term[of l \<sigma> vw, unfolded split]
  have "l \<cdot> \<sigma> =\<^sub>f (map_vars_term_mctxt vw C, map \<sigma> xs)" by simp
  also have "map \<sigma> xs = ts"
    by (rule nth_equalityI, insert len id, auto)
  finally show ?thesis .
qed

lemma split_vars_into_subst:
  assumes split: "split_vars l = (C,xs)"
    and len: "length ts = length xs"
    and id: "\<And> i. i < length xs \<Longrightarrow> \<sigma> (xs ! i) = ts ! i"
  shows "l \<cdot> \<sigma> =\<^sub>f (C,ts)"
  using split_vars_into_subst_map_vars_term[OF split len id, of "\<lambda> x. x"] by simp

lemma eqf_funas_term:
  "t =\<^sub>f (C,ss) \<Longrightarrow> funas_term t = funas_mctxt C \<union> \<Union>(funas_term ` set ss)"
proof (induct C arbitrary: t ss)
  case (MFun f Cs t ss)
  from eqf_MFunE[OF MFun(2)] obtain ts sss where
    t: "t = Fun f ts" and len: "length ts = length Cs" "length sss = length Cs" 
    and args: "\<And> i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)"
    and ss: "ss = concat sss"  by auto
  let ?lhs = "\<Union> {funas_term (ts ! i) | i. i < length Cs}"
  let ?f1 = "\<lambda> i. funas_mctxt (Cs ! i)"
  let ?f2 = "\<lambda> i. \<Union>(funas_term ` set (sss ! i))"
  let ?f = "\<lambda> i. ?f1 i \<union> ?f2 i"
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    note MFun(1)[OF mem args[OF i]]
  } note IH = this
  have "funas_term t = insert (f,length Cs) ?lhs"
    unfolding t using len by (auto simp: set_conv_nth)
  also have "?lhs = \<Union> {?f i | i. i < length Cs}" using IH by (metis (hide_lams)) 
  also have "\<dots> = \<Union> {?f1 i | i. i < length Cs} \<union> \<Union> {?f2 i | i. i < length Cs}" by auto
  also have "insert (f,length Cs) \<dots> = (insert (f,length Cs) (\<Union> {?f1 i | i. i < length Cs})) \<union> \<Union> {?f2 i | i. i < length Cs}" by auto
  also have "insert (f,length Cs) (\<Union> {?f1 i | i. i < length Cs}) = funas_mctxt (MFun f Cs)"
    by (auto simp: set_conv_nth)
  also have "\<Union> {?f2 i | i. i < length Cs} = \<Union>(funas_term ` set ss)" unfolding ss len(2)[symmetric] 
    using set_conv_nth[of sss] by auto
  finally show ?case .
next
  case MVar
  from eqfE[OF this]
  show ?case by auto
next
  case MHole
  from eqfE[OF this] show ?case by (cases ss, auto)
qed 

lemma eqf_all_ctxt_closed_step:
  assumes ctxt: "all_ctxt_closed F R"
    and ass: "t =\<^sub>f (D,ss)" "\<And> i. i < length ts \<Longrightarrow> (ss ! i, ts ! i) \<in> R" "length ss = length ts" "funas_term t \<subseteq> F"
      "UNION (set ts) funas_term \<subseteq> F"
  shows "(t, fill_holes D ts) \<in> R \<and> fill_holes D ts =\<^sub>f (D, ts)"
proof -
  have R: "(t,fill_holes D ts) \<in> R" using ass
  proof (induct D arbitrary: t ss ts)
    case (MVar x)
    from eqfE[OF MVar(1)] have id: "t = Var x" "ss = []" by auto
    with MVar(3) have ts: "ts = []" by auto
    show ?case unfolding id ts using all_ctxt_closed_sig_reflE[OF ctxt] by auto
  next
    case MHole
    from eqfE[OF MHole(1)] have id: "ss = [t]" by (cases ss, auto)
    with MHole(3) obtain t' where ts: "ts = [t']" by (cases ts, auto)
    show ?case using MHole(2)[of 0] unfolding id ts by auto
  next
    case (MFun f Cs)
    let ?fh = "fill_holes (MFun f Cs)"
    note eqft = MFun(2)
    from MFun(5)[unfolded eqf_funas_term[OF eqft]] MFun(6)
    have F: "(f, length Cs) \<in> F" "UNION (set Cs) funas_mctxt \<subseteq> F" "UNION (set ss) funas_term \<subseteq> F" "UNION (set ts) funas_term \<subseteq> F" by auto
    from eqfE[OF eqft] have t: "t = ?fh ss" and nh: "listsum (map num_holes Cs) = length ss" by auto    
    have length_upt: "\<And> x. length [0 ..< x] = x" by auto
    show ?case unfolding t fill_holes.simps
    proof (rule all_ctxt_closedE[OF ctxt], unfold length_map length_upt, rule F(1), simp)
      let ?list = "\<lambda> Cs ts i. (take (num_holes (Cs ! i)) (drop (listsum (map num_holes (take i Cs))) ts))"
      let ?fun = "\<lambda> Cs ts i. fill_holes (Cs ! i) (?list Cs ts i)"
      let ?val = "\<lambda> Cs ts i. map (?fun Cs ts) [0..<length Cs] ! i"
      fix i
      assume i: "i < length Cs"
      hence id: "?val Cs ts i = ?fun Cs ts i" "?val Cs ss i = ?fun Cs ss i" by auto
      from i have mem: "Cs ! i \<in> set Cs" by auto
      {
        fix j
        have "(j < length (?list Cs ts i) \<longrightarrow> (?list Cs ss i ! j, ?list Cs ts i ! j) \<in> R)
          \<and> num_holes (Cs ! i) = length (?list Cs ss i)" using MFun(3,4) nh i
        proof (induct Cs arbitrary: ts ss i j)
          case (Cons C Cs)
          let ?n = "num_holes C"
          from Cons(4) have len_ss: "length ss = ?n + listsum (map num_holes Cs)" by auto
          note len_ts = len_ss[unfolded Cons(3)] (* also for ts *)
          from len_ss len_ts have take: "length (take ?n ss) = length (take ?n ts)" by auto
          from len_ss have min_ss: "min (length ss) ?n = ?n" by auto
          from len_ts have min_ts: "min (length ts) ?n = ?n" by auto
          show ?case
          proof (cases i)
            case 0
            with Cons(2)[of j] show ?thesis by (simp add: min_ss)
          next
            case (Suc k)
            obtain ss' where ss: "ss = take ?n ss @ ss'" and ss': "ss' = drop ?n ss" by (metis append_take_drop_id)
            {
              fix n i
              have "drop (?n + n) ss = drop n ss'" "ss ! (?n + i) = ss' ! i" 
                unfolding arg_cong[OF ss, of "drop (?n + n)"] 
                unfolding arg_cong[OF ss, of nth] nth_append
                unfolding drop_append ss' by (auto simp: min_ss)
            } note id_ss = this
            from len_ts obtain ts' where ts: "ts = take ?n ts @ ts'" and ts': "ts' = drop ?n ts" by (metis append_take_drop_id)
            {
              fix n i
              have "drop (?n + n) ts = drop n ts'" "ts ! (?n + i) = ts' ! i"
                unfolding arg_cong[OF ts, of "drop (?n + n)"] 
                unfolding arg_cong[OF ts, of nth] nth_append
                unfolding drop_append ts' by (auto simp: min_ts)
            } note id_ts = this
            note IH = Cons(1)
            from Cons(3) have len: "length ss' = length ts'" unfolding ss' ts' by auto
            note IH = IH[OF _ this]
            from Cons(4) have nh: "listsum (map num_holes Cs) = length ss'" unfolding ss' by auto
            note IH = IH[OF _ this]
            from Cons(5) Suc have "k < length Cs" by auto
            note IH = IH[OF _ this]
            {
               fix i
               assume "i < length ts'"
               hence "?n + i < length ts" unfolding len_ts ts' nh len by auto
               from Cons(2)[OF this, unfolded id_ss id_ts]
               have "(ss' ! i, ts' ! i) \<in> R" .
            }
            note IH = IH[OF this, rule_format]
            let ?k = "listsum (map num_holes (take k Cs))"
            have "?thesis = ((j < length ts' - ?k \<and> j < num_holes (Cs ! k) \<longrightarrow>
                (drop ?k ss' ! j, drop ?k ts' ! j) \<in> R) \<and>
                num_holes (Cs ! k) = min (length ss' - ?k) (num_holes (Cs ! k)))" unfolding Suc by (simp add: id_ss id_ts)
            also have "..."
            proof (intro allI conjI impI, elim conjE)
              have "num_holes (Cs ! k) = length (take (num_holes (Cs ! k)) (drop ?k ss'))" using IH by blast
              also have "\<dots> = min (length ss' - ?k) (num_holes (Cs ! k))" by simp
              finally show "num_holes (Cs ! k) = min (length ss' - ?k) (num_holes (Cs ! k))" .
            next
              assume "j < length ts' - ?k" "j < num_holes (Cs ! k)"
              thus "(drop ?k ss' ! j, drop ?k ts' ! j) \<in> R"
                using IH[THEN conjunct1, THEN mp, of j] by auto
            qed
            finally show ?thesis .
          qed 
        qed auto
      }
      note main = this
      have CiF: "funas_mctxt (Cs ! i) \<subseteq> F" using F(2) i unfolding set_conv_nth by auto
      {
        have id: "?fun Cs ts i =\<^sub>f (Cs ! i, ?list Cs ts i)"
          by (rule eqfI, insert main MFun(4), auto)
        have "UNION (set (?list Cs ts i)) funas_term \<subseteq> UNION (set ts) funas_term" 
          by (rule UN_mono[OF subset_trans[OF set_take_subset set_drop_subset]], auto)
        hence ts_list: "UNION (set (?list Cs ts i)) funas_term \<subseteq> F" using F by auto
        have "funas_term (?fun Cs ts i) \<subseteq> funas_mctxt (Cs ! i) \<union> UNION (set (?list Cs ts i)) funas_term"
          unfolding eqf_funas_term[OF id] by auto
        also have "funas_mctxt (Cs ! i) \<subseteq> F" by (rule CiF)
        also have "UNION (set (?list Cs ts i)) funas_term \<subseteq> F" by fact
        finally have ts: "funas_term (?fun Cs ts i) \<subseteq> F" by blast
        have id: "?fun Cs ss i =\<^sub>f (Cs ! i, ?list Cs ss i)"
          by (rule eqfI, insert main MFun(4), auto)
        have "funas_term (?fun Cs ss i) \<subseteq> funas_mctxt (Cs ! i) \<union> UNION (set (?list Cs ss i)) funas_term"
          unfolding eqf_funas_term[OF id] by auto
        also have "funas_mctxt (Cs ! i) \<subseteq> F" by (rule CiF)
        also have "UNION (set (?list Cs ss i)) funas_term \<subseteq> UNION (set ss) funas_term" 
          by (rule UN_mono[OF subset_trans[OF set_take_subset set_drop_subset]], auto)
        also have "\<dots> \<subseteq> F" using F by auto
        finally have ss: "funas_term (?fun Cs ss i) \<subseteq> F" by blast
        note ts ss ts_list
      } note F_main = this
      have "((?val Cs ss i, ?val Cs ts i) \<in> R) = ((?fun Cs ss i, ?fun Cs ts i) \<in> R)" unfolding id ..
      also have "..." 
        by (rule MFun(1)[OF mem eqfI[OF refl] _ _ F_main(2-3)], insert main MFun(4), auto)
      finally show "(?val Cs ss i, ?val Cs ts i) \<in> R" .
      show "funas_term (?val Cs ss i) \<subseteq> F" unfolding id by fact
      show "funas_term (?val Cs ts i) \<subseteq> F" unfolding id by fact
    qed 
  qed
  show ?thesis
    by (rule conjI[OF R eqfI], insert eqfE[OF ass(1)] ass(3), auto)
qed

fun map_mctxt :: "('f \<Rightarrow> 'g) \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> ('g, 'v) mctxt"
where
  "map_mctxt _ (MVar x) = (MVar x)" |
  "map_mctxt _ (MHole) = MHole" |
  "map_mctxt fg (MFun f Cs) = MFun (fg f) (map (map_mctxt fg) Cs)"

fun ground_mctxt :: "('f, 'v) mctxt \<Rightarrow> bool"
where 
  "ground_mctxt (MVar _) = False" |
  "ground_mctxt MHole = True" |
  "ground_mctxt (MFun f Cs) = Ball (set Cs) ground_mctxt"

lemma ground_cap_till_funas [intro]:
  "ground_mctxt (cap_till_funas F t)"
  by (induct t) simp_all

lemma ground_eq_fill: "t =\<^sub>f (C,ss) \<Longrightarrow> ground t = (ground_mctxt C \<and> (\<forall> s \<in> set ss. ground s))" 
proof (induct C arbitrary: t ss)
  case (MVar x)
  from eqfE[OF this] show ?case by simp
next
  case (MHole t ss)
  from eqfE[OF this] show ?case by (cases ss, auto)
next
  case (MFun f Cs s ss)
  from eqf_MFunE[OF MFun(2)] obtain ts sss where s: "s = Fun f ts" and len: "length ts = length Cs" "length sss = length Cs" 
    and IH: "\<And> i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)" and ss: "ss = concat sss" by metis
  {
    fix i
    assume i: "i < length Cs"
    hence "Cs ! i \<in> set Cs" by simp
    from MFun(1)[OF this IH[OF i]]
    have "ground (ts ! i) = (ground_mctxt (Cs ! i) \<and> (\<forall>a\<in>set (sss ! i). ground a))" .
  } note IH = this
  note conv = set_conv_nth
  have "?case = ((\<forall>x\<in>set ts. ground x) = ((\<forall>x\<in>set Cs. ground_mctxt x) \<and> (\<forall>a\<in>set sss. \<forall>x\<in>set a. ground x)))"
    unfolding s ss by simp
  also have "..." unfolding conv[of ts] conv[of Cs] conv[of sss] len using IH by auto
  finally show ?case by simp
qed

lemma ground_fill_holes:
  assumes nh: "num_holes C = length ss"
  shows "ground (fill_holes C ss) = (ground_mctxt C \<and> (\<forall> s \<in> set ss. ground s))"
  by (rule ground_eq_fill[OF eqfI[OF refl nh]])

lemma split_vars_ground: "split_vars t = (C,xs) \<Longrightarrow> ground_mctxt C"
proof (induct t arbitrary: C xs)
  case (Fun f ts C xs)
  from Fun(2)[simplified] obtain Cs where C: "C = MFun f Cs" and Cs: "Cs = map (fst \<circ> split_vars) ts" by auto
  show ?case unfolding C ground_mctxt.simps
  proof
    fix C
    assume "C \<in> set Cs"
    from this[unfolded Cs] obtain t where t: "t \<in> set ts" and C: "C = fst (split_vars t)" unfolding o_def by auto
    from C obtain xs where split: "split_vars t = (C,xs)" by (cases "split_vars t", auto)
    show "ground_mctxt C"
      by (rule Fun(1)[OF t split]) 
  qed
qed auto

lemma split_vars_ground_vars:
  assumes "ground_mctxt C" and "num_holes C = length xs" 
  shows "split_vars (fill_holes C (map Var xs)) = (C, xs)"
using assms
proof (induct C arbitrary: xs)
  case (MHole xs)
  thus ?case by (cases xs, auto)
next
  case (MFun f Cs xs)
  have "fill_holes (MFun f Cs) (map Var xs) =\<^sub>f (MFun f Cs, map Var xs)"
    by (rule eqfI, insert MFun(3), auto)
  from eqf_MFunE[OF this] 
  obtain ts xss where fh: "fill_holes (MFun f Cs) (map Var xs) = Fun f ts"
    and lent: "length ts = length Cs"
    and lenx: "length xss = length Cs"
    and args: "\<And>i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, xss ! i)"
    and id: "map Var xs = concat xss" by auto
  from arg_cong[OF id, of "map the_Var"] have id2: "xs = concat (map (map the_Var) xss)" 
    by (auto simp: o_def map_concat)
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    with MFun(2) have ground: "ground_mctxt (Cs ! i)" by auto
    have "map Var (map the_Var (xss ! i)) = map id (xss ! i)" unfolding map_map o_def map_eq_conv
    proof
      fix x
      assume "x \<in> set (xss ! i)"
      with lenx i have "x \<in> set (concat xss)" by auto
      from this[unfolded id[symmetric]] show "Var (the_Var x) = id x" by auto
    qed
    hence idxss: "map Var (map the_Var (xss ! i)) = xss ! i" by auto
    note rec = eqfE[OF args[OF i]]
    note IH = MFun(1)[OF mem ground, of "map the_Var (xss ! i)", unfolded rec(2) idxss rec(1)[symmetric]]
    from IH have "split_vars (ts ! i) = (Cs ! i, map the_Var (xss ! i))" by auto
    note this idxss
  }
  note IH = this
  have "?case = (map fst (map split_vars ts) = Cs \<and> concat (map snd (map split_vars ts)) = concat (map (map the_Var) xss))"
    unfolding fh unfolding id2 by auto
  also have "\<dots>"
  proof (rule conjI[OF nth_equalityI[OF _ allI[OF impI]] arg_cong[of _ _ concat, OF nth_equalityI, rule_format]], unfold length_map lent lenx)
    fix i
    assume i: "i < length Cs" 
    with arg_cong[OF IH(2)[OF this], of "map the_Var"]
      IH[OF this] show "map snd (map split_vars ts) ! i = map (map the_Var) xss ! i" using lent lenx by auto
  qed (insert IH lent, auto)
  finally show ?case .
qed auto

lemma ground_map_mctxt[simp]: "ground_mctxt (map_mctxt fg C) = ground_mctxt C"
  by (induct C, auto)

lemma num_holes_map_mctxt[simp]: "num_holes (map_mctxt fg C) = num_holes C"
proof (induct C)
  case (MFun f Cs)
  thus ?case by (induct Cs, auto)
qed auto

lemma split_vars_map_mctxt:
  assumes split: "split_vars t = (map_mctxt fg C, xs)"
  shows "split_vars (fill_holes C (map Var xs)) = (C, xs)"
proof -
  from split_vars_ground[OF split] have ground: "ground_mctxt C" by simp
  from split_vars_num_holes[of t, unfolded split] have nh: "num_holes C = length xs" by auto
  show ?thesis
    by (rule split_vars_ground_vars[OF ground nh])
qed

lemma subst_eq_map_decomp:
  assumes "t \<cdot> \<sigma> = map_funs_term fg s"
  shows "\<exists> C xs \<delta>s. s =\<^sub>f (C,\<delta>s) \<and> split_vars t = (map_mctxt fg C, xs) \<and> (\<forall> i < length xs. 
    \<sigma> (xs ! i) = map_funs_term fg (\<delta>s ! i))" 
using assms
proof (induct t arbitrary: s)
  case (Var x s)
  show ?case
    by (rule exI[of _ MHole], rule exI[of _ "[x]"], rule exI[of _ "[s]"], insert Var, auto)
next
  case (Fun g ts s)
  from Fun(2) obtain f ss where s: "s = Fun f ss" and g: "g = fg f" by (cases s, auto)
  from Fun(2)[unfolded s] have id: "map (\<lambda> t. t \<cdot> \<sigma>) ts = map (map_funs_term fg) ss" by auto
  from arg_cong[OF this, of length] have len: "length ts = length ss" by auto
  from map_nth_conv[OF id] have args: "\<And> i. i < length ts \<Longrightarrow> ts ! i \<cdot> \<sigma> = map_funs_term fg (ss ! i)" by auto
  let ?P = "\<lambda> C xs \<delta>s i. ss ! i =\<^sub>f (C, \<delta>s) \<and>
          split_vars (ts ! i) = (map_mctxt fg C, xs) \<and>
          (\<forall>i<length xs. \<sigma> (xs ! i) = map_funs_term fg (\<delta>s ! i))"
  {
    fix i
    assume i: "i < length ts"
    hence mem: "ts ! i \<in> set ts" by auto
    note IH = Fun(1)[OF this args[OF i]]
  }
  hence "\<forall> i. \<exists> C xs \<delta>s. i < length ts \<longrightarrow> ?P C xs \<delta>s i" by blast
  from choice[OF this] obtain Cs where "\<forall> i. \<exists> xs \<delta>s. i < length ts \<longrightarrow> ?P (Cs i) xs \<delta>s i" by blast
  from choice[OF this] obtain xss where "\<forall> i. \<exists> \<delta>s. i < length ts \<longrightarrow> ?P (Cs i) (xss i) \<delta>s i" by blast
  from choice[OF this] obtain \<delta>ss where IH: "\<And> i. i < length ts \<Longrightarrow> ?P (Cs i) (xss i) (\<delta>ss i) i" by blast
  let ?n = "[0 ..< length ts]"
  let ?Cs = "map Cs ?n"
  let ?C = "MFun f ?Cs"
  let ?xs = "concat (map xss ?n)"
  let ?\<delta>s = "concat (map \<delta>ss ?n)"
  let ?g = "fg f"
  show ?case unfolding s g 
  proof (rule exI[of _ ?C], rule exI[of _ ?xs], rule exI[of _ ?\<delta>s], intro conjI)
    show "Fun f ss =\<^sub>f (?C, ?\<delta>s)" 
      by (rule eqf_MFunI, insert IH len, auto)
  next
    have 
      "(split_vars (Fun ?g ts) = (map_mctxt fg ?C, ?xs)) 
        = (map (fst \<circ> split_vars) ts = map (map_mctxt fg \<circ> Cs) [0..<length ss] 
          \<and>concat (map (snd \<circ> split_vars) ts) = ?xs)" 
      (is "?goal = _")
      using len by auto
    also have "\<dots>"
      by (rule conjI[OF nth_map_conv arg_cong[of _ _ concat, OF nth_equalityI]], insert IH len, auto)
    finally show ?goal .
  next
    show "\<forall> i < length ?xs. \<sigma> (?xs ! i) = map_funs_term fg (?\<delta>s ! i)"
    proof (rule concat_all_nth, simp, unfold length_map length_upt)
      fix i
      assume "i < length ts - 0"
      hence i: "i < length ts" by auto
      from IH[OF i] have "split_vars (ts ! i) = (map_mctxt fg (Cs i), xss i)" by blast
      from split_vars_map_mctxt[OF this] split_vars_num_holes[of "fill_holes (Cs i) (map Var (xss i))"]
      have len: "length (xss i) = num_holes (Cs i)" by simp
      also have "\<dots> = length (\<delta>ss i)" by (rule eqfE(2), insert IH[OF i], auto)
      finally 
      show "length (map xss [0..<length ts] ! i) = length (map \<delta>ss [0..<length ts] ! i)" using i by auto
    qed (insert IH, auto)
  qed
qed

lemma map_funs_term_fill_holes:
  "num_holes C = length ss \<Longrightarrow>
    map_funs_term fg (fill_holes C ss) =\<^sub>f (map_mctxt fg C, map (map_funs_term fg) ss)"
proof (induct C arbitrary: ss)
  case (MHole ss)
  thus ?case by (cases ss, auto)
next
  case MVar thus ?case by auto
next
  case (MFun f Cs ss)
  from MFun(2) have "fill_holes (MFun f Cs) ss =\<^sub>f (MFun f Cs, ss)" by auto
  from eqf_MFunE[OF this] obtain ts sss where fh: "fill_holes (MFun f Cs) ss = Fun f ts"
    and lts: "length ts = length Cs"
    and lsss: "length sss = length Cs"
    and args: "\<And>i. i < length Cs \<Longrightarrow> ts ! i =\<^sub>f (Cs ! i, sss ! i)"
    and sss: "ss = concat sss" by auto
  {
    fix i
    assume i: "i < length Cs"
    hence mem: "Cs ! i \<in> set Cs" by auto
    from MFun(1)[OF mem]  eqfE[OF args[OF i]] have 
      "map_funs_term fg (ts ! i) =\<^sub>f (map_mctxt fg (Cs ! i), map (map_funs_term fg) (sss ! i))" by auto
  } note IH = this
  show ?case unfolding fh 
    unfolding map_mctxt.simps sss map_concat term.simps
  proof (rule eqf_MFunI, unfold length_map, simp add: lsss, simp add: lts)
    fix i
    assume i: "i < length Cs"
    have "(map (map_funs_term fg) ts ! i =\<^sub>f (map (map_mctxt fg) Cs ! i, map (map (map_funs_term fg)) sss ! i)) =
      (map_funs_term fg (ts ! i) =\<^sub>f (map_mctxt fg (Cs ! i), map (map_funs_term fg) (sss ! i)))" (is "?goal = _")
      using i lts lsss by auto
    also have "\<dots>" by (rule IH[OF i])
    finally show ?goal .
  qed
qed

lemma eqf_imp_subt:
  assumes s: "s =\<^sub>f (C,ts)"
  and t: "t \<in> set ts"
  shows "s \<unrhd> t"
proof -
  from t obtain bef aft where ts: "ts = bef @ t # aft"
    by (metis split_list)
  note s = eqfE[OF s[unfolded ts], simplified]
  from fill_holes_ctxt_main[OF s(2)] obtain D where "fill_holes C (bef @ t # aft) = D\<langle>t\<rangle>" by auto
  from this[folded s(1)] show ?thesis by auto
qed

fun poss_mctxt :: "('f, 'v) mctxt \<Rightarrow> pos set"
where
  "poss_mctxt (MVar x) = {\<epsilon>}" |
  "poss_mctxt MHole = {}" |
  "poss_mctxt (MFun f cs) = {\<epsilon>} \<union> \<Union>(set (map (\<lambda> i. (\<lambda> p. i <# p) ` poss_mctxt (cs ! i)) [0 ..< length cs]))"

lemma poss_mctxt_simp [simp]:
  "poss_mctxt (MFun f cs) = {\<epsilon>} \<union> {i <# p | i p. i < length cs \<and> p \<in> poss_mctxt (cs ! i)}"
  by auto
declare poss_mctxt.simps(3)[simp del]

fun hole_poss :: "('f, 'v) mctxt \<Rightarrow> pos set"
where
  "hole_poss (MVar x) = {}" |
  "hole_poss MHole = {\<epsilon>}" |
  "hole_poss (MFun f cs) = \<Union>(set (map (\<lambda> i. (\<lambda> p. i <# p) ` hole_poss (cs ! i)) [0 ..< length cs]))"

lemma hole_poss_simp [simp]:
  "hole_poss (MFun f cs) = {i <# p | i p. i < length cs \<and> p \<in> hole_poss (cs ! i)}"
  by auto
declare hole_poss.simps(3)[simp del]

lemma mctxt_of_term_fill_holes [simp]:
  "fill_holes (mctxt_of_term t) [] = t"
proof (induct t)
  case (Fun f ts) 
  then have "fill_holes (mctxt_of_term (Fun f ts)) [] = Fun f (map (\<lambda>i. (ts!i)) [0..<length ts])"
    unfolding mctxt_of_term.simps partition_holes_fill_holes_conv partition_by_Nil map_map by auto
  also have "... = Fun f ts" using map_nth by auto
  ultimately show ?case by auto
qed (auto)

fun
  subst_apply_mctxt :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v, 'w) gsubst \<Rightarrow> ('f, 'w) mctxt" (infixl "\<cdot>mc" 67)
where
  "MHole \<cdot>mc _ = MHole" |
  "(MVar x) \<cdot>mc \<sigma> = mctxt_of_term (\<sigma> x)" |
  "(MFun f cs) \<cdot>mc \<sigma> = MFun f [c \<cdot>mc \<sigma> . c \<leftarrow> cs]"

lemma subst_apply_mctxt_numholes: 
  shows "num_holes (c \<cdot>mc \<sigma>) = num_holes c"
proof (induct c arbitrary: \<sigma>) 
  case (MFun f cs)
  have "num_holes (MFun f cs \<cdot>mc \<sigma>) = listsum [num_holes (c \<cdot>mc \<sigma>) . c \<leftarrow> cs]"
    unfolding subst_apply_mctxt.simps num_holes.simps map_map comp_def by auto
  also have "... = listsum [num_holes c . c \<leftarrow> cs]" using MFun(1) 
    by (metis (lifting, no_types) map_cong)
  ultimately show ?case by auto
qed (auto)

lemma subst_apply_mctxt_fill_holes: 
  assumes nh: "num_holes c = length ts"
  shows "(fill_holes c ts) \<cdot> \<sigma> = fill_holes (c \<cdot>mc \<sigma>) [ ti \<cdot> \<sigma> . ti \<leftarrow> ts]"
using nh
proof (induct c arbitrary: ts)
  case MHole 
  then obtain t where ts: "ts = [t]" 
    unfolding num_holes.simps unfolding One_nat_def using Suc_length_conv length_0_conv by metis
  show ?case unfolding ts by simp
next 
  case (MVar x)
  then have ts: "ts = []" using length_0_conv by auto 
  show ?case unfolding ts by auto
next
  case (MFun f cs) 
  note IH = MFun(1)
  note nh = MFun(2)[unfolded num_holes.simps]
  let ?c = "MFun f cs"
  let ?cs\<sigma> = "map (\<lambda>c. c \<cdot>mc \<sigma>) cs"

  {
    fix i
    assume i: "i < length cs"
    have nh_map: "\<And> j. j < length cs \<Longrightarrow> num_holes (cs!j) = num_holes (?cs\<sigma> ! j)" 
      using nth_map subst_apply_mctxt_numholes by metis

    have "fill_holes (cs ! i) (partition_holes ts cs ! i) \<cdot> \<sigma> = 
      fill_holes ((cs ! i) \<cdot>mc \<sigma>) (partition_holes [ ti \<cdot> \<sigma> . ti \<leftarrow> ts] cs ! i)"
      using IH [OF nth_mem [OF i]] and nh and i by auto
    also have "... = fill_holes (?cs\<sigma> ! i) (partition_holes [ti \<cdot> \<sigma> . ti \<leftarrow> ts] ?cs\<sigma> ! i)"
      unfolding nth_map[OF i] using partition_holes_map_ctxt[OF _ nh_map] length_map by metis
    ultimately have "fill_holes (cs ! i) (partition_holes ts cs ! i) \<cdot> \<sigma> = fill_holes (?cs\<sigma> ! i) (partition_holes [ti \<cdot> \<sigma> . ti \<leftarrow> ts] ?cs\<sigma> ! i)"
      by auto
    } note ith = this

  have "fill_holes ?c ts \<cdot> \<sigma> = Fun f [fill_holes (?cs\<sigma> ! i) (partition_holes [ ti \<cdot> \<sigma> . ti \<leftarrow> ts] ?cs\<sigma> ! i) . i \<leftarrow> [0..<length cs]]"  
  unfolding partition_holes_fill_holes_conv map_map using ith using comp_def  by auto 
  also have "... = fill_holes (?c \<cdot>mc \<sigma>) [ ti \<cdot> \<sigma> . ti \<leftarrow> ts]"  
    unfolding subst_apply_mctxt.simps partition_holes_fill_holes_conv length_map ..
  ultimately show ?case by auto
qed
  
lemma subst_apply_mctxt_sound: 
  assumes "t =\<^sub>f (c,ts)"
  shows "t \<cdot> \<sigma> =\<^sub>f (c \<cdot>mc \<sigma>, [ ti \<cdot> \<sigma> . ti \<leftarrow> ts])"
proof (rule eqfI, insert subst_apply_mctxt_numholes subst_apply_mctxt_fill_holes[OF eqfE(2)[OF assms]] eqfE[OF assms] eqfE(2)[OF assms, symmetric], auto) qed

fun fill_holes_mctxt :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt list \<Rightarrow> ('f, 'v) mctxt"
where
  "fill_holes_mctxt (MVar x) _ = MVar x" |
  "fill_holes_mctxt MHole [] = MHole" |
  "fill_holes_mctxt MHole [t] = t" |
  "fill_holes_mctxt (MFun f cs) ts = (MFun f (map (\<lambda> i. fill_holes_mctxt (cs ! i) 
    (take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ts))) [0 ..< length cs]))"

lemma fill_holes_mctxt_MHole [simp]:
  "length ts = Suc 0 \<Longrightarrow> fill_holes_mctxt MHole ts = hd ts"
  by (cases ts) simp_all

lemma partition_holes_fill_holes_mctxt_conv:
  "fill_holes_mctxt (MFun f Cs) ts =
    MFun f [fill_holes_mctxt (Cs ! i) (partition_holes ts Cs ! i). i \<leftarrow> [0 ..< length Cs]]"
  by (simp add: partition_by_nth take_map)

lemma partition_holes_fill_holes_mctxt_conv':
  "fill_holes_mctxt (MFun f Cs) ts =
    MFun f (map (split fill_holes_mctxt) (zip Cs (partition_holes ts Cs)))"
  unfolding zip_nth_conv [of Cs "partition_holes ts Cs", simplified]
    and partition_holes_fill_holes_mctxt_conv by simp

lemma partition_holes_fill_holes_conv':
  "fill_holes (MFun f Cs) ts =
    Fun f (map (split fill_holes) (zip Cs (partition_holes ts Cs)))"
  unfolding zip_nth_conv [of Cs "partition_holes ts Cs", simplified]
    and partition_holes_fill_holes_conv by simp

lemma fill_holes_mctxt_MFun_replicate_length [simp]:
  "fill_holes_mctxt (MFun c (replicate (length Cs) MHole)) Cs = MFun c Cs"
  unfolding partition_holes_fill_holes_mctxt_conv'
  by (induct Cs) simp_all

lemma fill_holes_MFun_replicate_length [simp]:
  "fill_holes (MFun c (replicate (length ts) MHole)) ts = Fun c ts"
  unfolding partition_holes_fill_holes_conv'
  by (induct ts) simp_all

lemma funas_mctxt_fill_holes_mctxt [simp]:
  assumes "num_holes C = length Ds"
  shows "funas_mctxt (fill_holes_mctxt C Ds) = funas_mctxt C \<union> \<Union>(set (map funas_mctxt Ds))"
  (is "?f C Ds = ?g C Ds")
using assms
proof (induct C arbitrary: Ds)
  case MHole
  then show ?case by (cases Ds) simp_all
next
  case (MFun f Cs)
  then have num_holes: "listsum (map num_holes Cs) = length Ds" by simp
  let ?ys = "partition_holes Ds Cs"
  have "\<And>i. i < length Cs \<Longrightarrow> ?f (Cs ! i) (?ys ! i) = ?g (Cs ! i) (?ys ! i)"
    using MFun by (metis nth_mem num_holes.simps(3) length_partition_holes_nth)
  then have "(\<Union>i \<in> {0 ..< length Cs}. ?f (Cs ! i) (?ys ! i)) =
    (\<Union>i \<in> {0 ..< length Cs}. ?g (Cs ! i) (?ys ! i))" by simp
  then show ?case
    using num_holes
    unfolding partition_holes_fill_holes_mctxt_conv
    by (simp add: UN_Un_distrib UN_upt_len_conv [of _ _ "\<lambda>x. \<Union>(set x)"] UN_set_partition_by_map)
qed simp

lemma fill_holes_mctxt_MFun:
  assumes lCs: "length Cs = length ts"
    and lss: "length ss = length ts"
    and rec: "\<And> i. i < length ts \<Longrightarrow> num_holes (Cs ! i) = length (ss ! i) \<and> fill_holes_mctxt (Cs ! i) (ss ! i) = ts ! i"
  shows "fill_holes_mctxt (MFun f Cs) (concat ss) = MFun f ts" 
  unfolding fill_holes_mctxt.simps mctxt.simps
    by (rule conjI[OF refl], rule fill_holes_arbitrary[OF lCs lss rec])

lemma num_holes_fill_holes_mctxt: 
  assumes "num_holes C = length Ds"
  shows "num_holes (fill_holes_mctxt C Ds) = listsum (map num_holes Ds)"
using assms
proof (induct C arbitrary: Ds)
  case MHole
  then show ?case by (cases Ds) simp_all
next
  case (MFun f Cs)
  then have *: "map (num_holes \<circ> (\<lambda>i. fill_holes_mctxt (Cs ! i) (partition_holes Ds Cs ! i))) [0..<length Cs] =
    map (\<lambda>i. listsum (map num_holes (partition_holes Ds Cs ! i))) [0 ..< length Cs]"
    and "listsum (map num_holes Cs) = length Ds"
    by simp_all
  then show ?case
    using map_upt_len_conv [of "\<lambda>x. listsum (map num_holes x)" "partition_holes Ds Cs"]
    unfolding partition_holes_fill_holes_mctxt_conv by (simp add: *)
qed simp

lemma fill_holes_mctxt_fill_holes: 
  assumes len_ds: "length ds = num_holes c"
    and nh: "num_holes (fill_holes_mctxt c ds) = length ss"
  shows "fill_holes (fill_holes_mctxt c ds) ss =
    fill_holes c [fill_holes (ds ! i) (partition_holes ss ds ! i). i \<leftarrow> [0 ..< num_holes c]]"
using nh len_ds proof (induct c arbitrary: ss ds)
  case (MVar x) then have ds: "ds = []"  by auto
  show ?case unfolding ds by simp
next
  case (MHole) then
  obtain d where ds: "ds = [d]" 
    unfolding num_holes.simps
    using Suc_length_conv length_0_conv One_nat_def by metis
  have "length ss = num_holes d"
    unfolding MHole(1)[symmetric] ds by simp
  then show ?case unfolding ds partition_holes_step by auto
next 
  case (MFun f cs)
  note IH=MFun(1)
  let ?c = "MFun f cs"

  def csi \<equiv> "\<lambda> j . fill_holes_mctxt (cs ! j) (partition_holes ds cs ! j)"
  def ssi \<equiv> "\<lambda> j . partition_holes ss (map csi [0..<length cs]) ! j"

  {
    fix i 
    assume i: "i < length cs"
    hence i': "i < length [0..<length cs]" by auto
    hence i'': "[0..<length cs] ! i = i" by auto
    have i_ph: "i < length (partition_holes ds cs)" using i by simp

    let ?dsis = "partition_holes ds cs ! i"
    let ?td_l = "\<lambda> ls . take (num_holes (cs ! i)) (drop (listsum (map num_holes (take i cs))) ls)"    
    let ?td_nh = "\<lambda> ls . take (length ?dsis) (drop (listsum (map length (take i (partition_holes ds cs)))) ls)"

    have l_dsi: "\<And>  j. j < length cs \<Longrightarrow> length (partition_holes ds cs ! j) = num_holes (cs ! j)" 
      using MFun.prems by simp
    have l_nh_conv: "map length (partition_holes ds cs) = map num_holes cs"
      using l_dsi by (simp add: map_nth_eq_conv)

    { 
      fix ls
      have "?td_l ls = ?td_nh ls" unfolding l_dsi[OF i] unfolding take_map[symmetric] l_nh_conv by auto
    } note td_l_nh_conv = this
      
    have ph_ds_i: "partition_holes ds cs ! i = ?td_l ds"
      using partition_by_nth [of i "map num_holes cs" ds] and i by (simp add: take_map)
    have ph_ss_i: "partition_holes (ssi i) (partition_holes ds cs ! i) = ?td_l (partition_holes ss ds)" 
    proof -
      from i have tk: "take i [0..<length cs] = [0..<i]" using take_upt by auto
      have nh_csi: "\<And>  j. j < length cs \<Longrightarrow> num_holes (csi j) = listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! j]"
        unfolding csi_def using num_holes_fill_holes_mctxt [OF l_dsi [symmetric]] i  by auto
      have nh_csjs: "[num_holes (csi j) . j \<leftarrow> [0..<i]] = [listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! j] . j \<leftarrow> [0..<i]]"
        using nh_csi i by auto 
      have i_csis: "i < length (map csi [0..<length cs])" unfolding length_map using i by auto
      then have *: "i < length (map num_holes (map csi [0 ..< length cs]))" by simp
      have ssi_take_drop: 
        "ssi i = take (listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! i]) 
        (drop (listsum [listsum [num_holes di . di \<leftarrow> partition_holes ds cs ! j] . j \<leftarrow> [0..<i]]) ss)"
        unfolding ssi_def
        unfolding partition_by_nth [OF *, of ss]
        unfolding take_map tk map_map o_def
        unfolding nth_map [OF i', of "\<lambda>x. num_holes (csi x)"] i'' nh_csi [OF i] nh_csjs ..
      
      {
        fix j
        assume j: "j < length (partition_holes ds cs ! i)"

        let ?a = "map num_holes ?dsis"
        let ?l = "map num_holes (take j ?dsis)"
        let ?r = "map num_holes (drop j ?dsis)"


        have "listsum ?a - listsum ?l = listsum (?l @ ?r) - listsum ?l"
          using append_take_drop_id[symmetric] take_map drop_map by metis
        then have ls_r: "listsum ?a - listsum ?l = listsum  ?r" by auto

        from j obtain di dis where "drop j ?dsis = di # dis"
          by (metis Cons_nth_drop_Suc)
        moreover have "?dsis ! j = drop j ?dsis ! 0"
          using j nth_drop by auto
        ultimately have "num_holes (?dsis ! j) \<le> listsum ?r" proof (cases "?r", auto) qed
        then have "min (num_holes (?dsis ! j)) (listsum ?a - listsum ?l) = num_holes (?dsis ! j)" using ls_r by auto
      } note takes = this

      have t_i_suci_conv: "take i (map num_holes cs) @ (num_holes (cs ! i) # drop (Suc i) (map num_holes cs)) = map num_holes cs"
        by (metis length_map nth_map id_take_nth_drop i)
      have "listsum (map length (take i (partition_holes ds cs))) + length (partition_holes ds cs ! i) 
            \<le> listsum (take i (map num_holes cs) @ (num_holes (cs ! i) # drop (Suc i) (map num_holes cs)))"
        unfolding take_map[symmetric] l_nh_conv l_dsi[OF i] listsum_append listsum_simps(2) by auto
      also have "... = length ds"
        unfolding MFun(3) t_i_suci_conv by simp
      ultimately have ti_le_len_ds: "listsum (map length (take i (partition_holes ds cs))) + length (partition_holes ds cs ! i) \<le> length ds"
        by auto
      then have 
        "?td_nh [0..<length ds] 
         = [(listsum (map length (take i (partition_holes ds cs)))) ..< (listsum (map length (take i (partition_holes ds cs)))) + (length ?dsis)]"
        unfolding drop_upt using take_upt by auto
      also have "... = [ id (listsum (map length (take i (partition_holes ds cs))) + j). j \<leftarrow> [0..<length ?dsis]]" 
        unfolding map_upt_add'[symmetric] by auto
      ultimately have td_nh_ixs: 
        "?td_nh [0..<length ds] = [ j + listsum (map length (take i (partition_holes ds cs))). j \<leftarrow> [0..<length ?dsis]]" by auto

      {
        fix j
        assume j: "j < length ?dsis"
        then have b: "(listsum (take i (map length (partition_holes ds cs)))) + j \<le> length ds" 
          unfolding take_map using ti_le_len_ds by auto

        have "ds ! (j + listsum (map length (take i (partition_holes ds cs)))) = ?dsis ! j"
          unfolding partition_holes_nth [OF i] 
          unfolding l_dsi [OF i, symmetric] nth_take [OF j] take_map [symmetric]
          unfolding nth_drop [OF b, symmetric, unfolded add.commute]
          unfolding l_nh_conv ..
      } note dsis_j_conv = this

      {
        fix j
        assume j: "j < length ?dsis"
        then have min: "min j (num_holes (cs!i)) = j" 
          by (metis l_dsi[OF i] less_not_sym min_less_iff_disj nat_less_cases)
        let ?s = "listsum (map length (take i (partition_holes ds cs)))"

        have dsi_j: "take j (drop ?s ds) = take j ?dsis"
          unfolding partition_holes_nth [OF i] take_take take_map [symmetric] l_nh_conv nth_map [OF i] min ..
        
        have map_l_ph_conv: "map length (take i (partition_holes ds cs)) = map num_holes (take i cs)" 
          unfolding take_map[symmetric] l_nh_conv ..

        from i have i_leq: "i \<le> length cs" by auto
        have dsjs: 
          "take ?s ds = concat [ partition_holes ds cs ! j. j \<leftarrow> [0..<i]]"
          unfolding map_l_ph_conv take_upt_idx [OF i] map_map o_def
          unfolding partition_holes_nth [OF i, of ds]
          unfolding map_l_ph_conv concat_partition_holes_upt[OF i_leq] 
          unfolding take_upt_idx[OF i] map_map o_def ..

        have conv:
          "take (j + ?s) ds = concat [ partition_holes ds cs ! j. j \<leftarrow> [0..<i]] @ take j ?dsis" 
          using take_add add.commute dsi_j dsjs by metis
        
        have 
          "listsum (map num_holes (take (j + listsum (map length (take i (partition_holes ds cs)))) ds))
          = listsum (map num_holes (take j ?dsis))
          + (\<Sum>k\<leftarrow>[0..<i]. listsum (map num_holes (partition_holes ds cs ! k)))"
          unfolding conv by (simp add: map_concat o_def listsum_concat)
      } note dsis_j_conv2 = this
      
      have "?td_l (partition_holes ss ds)
        = [take (num_holes (ds ! i)) (drop (listsum (map num_holes (take i ds))) ss) . i \<leftarrow> ?td_nh [0..<length ds]]"
        unfolding td_l_nh_conv
        unfolding td_l_nh_conv partition_holes_def drop_map take_map ..
      also have 
        "... 
        = [ take (num_holes (?dsis ! j)) 
        (drop (listsum (map num_holes (take j ?dsis))
        + (\<Sum>k\<leftarrow>[0..<i]. listsum (map num_holes (partition_holes ds cs ! k)))) ss)
        . j \<leftarrow> [0..<length ?dsis]]"
        unfolding td_nh_ixs map_map o_def using dsis_j_conv dsis_j_conv2 by auto
      also have 
        "... = partition_holes (ssi i) ?dsis"
        unfolding partition_holes_def[of "(ssi i)"] 
        unfolding ssi_take_drop drop_take take_take drop_drop
        using takes by auto
      ultimately show ?thesis by auto
    qed
    
    have l1: "length (?td_l ds) = num_holes (cs ! i)" by (metis l_dsi[OF i] ph_ds_i)
    have l2: "length (?td_l (partition_holes ss ds)) = num_holes (cs ! i)" using l1 by simp
    from l1 l2 have l3: "length (zip (?td_l ds) (?td_l (partition_holes ss ds))) = num_holes (cs ! i)" by auto

    have nh_i: "num_holes (csi i) = length (ssi i)" 
      unfolding csi_def ssi_def 
      using length_partition_holes_nth [OF MFun.prems(1)[unfolded partition_holes_fill_holes_mctxt_conv num_holes.simps], unfolded  length_map]
      by (metis (lifting, no_types) i' i'' nth_map)

    have "[ fill_holes (partition_holes ds cs ! i ! j) (partition_holes (ssi i) (partition_holes ds cs ! i) ! j) . j \<leftarrow> [0..<num_holes (cs ! i)] ]
         = map (\<lambda> j . fill_holes (?td_l ds ! j) (?td_l (partition_holes ss ds) ! j)) [0..<num_holes (cs ! i)]"
      unfolding ph_ss_i using ph_ds_i by auto
    also have 
      "... = map (\<lambda> j . (\<lambda> e. fill_holes (fst e) (snd e)) (?td_l ds ! j, ?td_l (partition_holes ss ds) ! j)) [0..<num_holes (cs ! i)]"
      by auto
    also have 
      "... = map (\<lambda> j . (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds)) ! j)) [0..<num_holes (cs ! i)]" 
      using nth_zip[symmetric] l1 l2 by simp
    also have 
      "... = map (\<lambda> j . map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds))) ! j) [0..<num_holes (cs ! i)]" 
      using l3 nth_map[symmetric] by simp
    also have 
      "... = map (\<lambda> j . map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds))) ! j) 
                 [0..<length (map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds))))]"
      unfolding length_map using l3 by auto
    also have 
      "... = map (\<lambda> e. fill_holes (fst e) (snd e)) (zip (?td_l ds) (?td_l (partition_holes ss ds)))" 
      using map_nth by metis 
    also have 
      "... = map (\<lambda> e. fill_holes (fst e) (snd e)) (?td_l (zip ds (partition_holes ss ds)))" 
      unfolding take_zip[symmetric] drop_zip[symmetric] ..
    also have
      "... = map (\<lambda> e. fill_holes (fst e) (snd e)) (partition_holes (zip ds (partition_holes ss ds)) cs ! i)"
      unfolding partition_holes_nth [OF i] ..
    also have
      "... = partition_holes (map (\<lambda>e. fill_holes (fst e) (snd e)) (zip ds (partition_holes ss ds))) cs ! i" 
      using i by simp
    also have
      "... = partition_holes (map (\<lambda>k. fill_holes (ds ! k) (partition_holes ss ds ! k)) [0..<num_holes ?c]) cs ! i"
      unfolding zip_nth_conv[OF length_partition_holes[symmetric]] map_map o_def length_partition_holes MFun.prems(2) by auto
    ultimately 
    have fh: 
      "(map (\<lambda>j. fill_holes (partition_holes ds cs ! i ! j) (partition_holes (ssi i) (partition_holes ds cs ! i) ! j))
            [0..<num_holes (cs ! i)]) 
       = partition_holes (map (\<lambda>k. fill_holes (ds ! k) (partition_holes ss ds ! k)) [0..<num_holes ?c]) cs ! i"
      by auto
    have 
      "fill_holes (map csi [0..<length cs] ! i) (ssi i) = fill_holes (csi i) (ssi i)" (is "?l = _ ") 
      unfolding csi_def ssi_def  using nth_map[OF i'] i'' by metis
    also have 
      "... = fill_holes (cs!i) 
                        (map (\<lambda>j. fill_holes (partition_holes ds cs ! i ! j) 
                                             (partition_holes (ssi i) (partition_holes ds cs ! i) ! j)) 
                             [0..<num_holes (cs ! i)])" 
      using IH[OF nth_mem[OF i] nh_i[unfolded csi_def ssi_def] l_dsi[OF i]] 
      unfolding csi_def ssi_def by auto
    also have 
      "... = fill_holes (cs!i) 
                        (partition_holes (map (\<lambda>j. fill_holes (ds ! j) (partition_holes ss ds ! j)) [0..<num_holes ?c]) cs ! i)" (is "_ = ?r") 
      using fh by auto
    ultimately have "?l = ?r" by auto
  }
  then show ?case 
    unfolding 
      ssi_def csi_def
      partition_holes_fill_holes_mctxt_conv partition_holes_fill_holes_conv length_map  by auto
qed

lemma fill_holes_mctxt_sound:
  assumes len_ds: "length ds = num_holes c"
  and len_sss: "length sss = num_holes c"
  and len_ts: "length ts = num_holes c"
  and insts: "\<And> i. i < length ds \<Longrightarrow> ts!i =\<^sub>f (ds!i, sss!i)"
  shows "fill_holes c ts =\<^sub>f (fill_holes_mctxt c ds, concat sss)"
proof (rule eqfI)
  note l_nh_i = eqfE(2)[OF insts]

  from partition_holes_concat_id[OF _ l_nh_i] len_ds len_sss
  have concat_sss: "partition_holes (concat sss) ds = sss" by auto

  then show nh: "num_holes (fill_holes_mctxt c ds) = length (concat sss)"
    unfolding num_holes_fill_holes_mctxt [OF len_ds [symmetric]] length_concat
    by (metis l_nh_i len_ds len_sss nth_map_conv)  

  have ts: "ts = [fill_holes (ds ! i) (partition_holes (concat sss) ds ! i) . i \<leftarrow> [0..<num_holes c]]" (is "_ = ?fhs")
  proof (rule nth_equalityI)
    show l_fhs: "length ts = length ?fhs" unfolding length_map
      by (metis diff_zero len_ts length_upt)
    {
      fix i
      assume i: "i < length ts"
      then have i': "i < length [0..<num_holes c]" 
        by (metis diff_zero len_ts length_upt)
      have "ts!i = ?fhs ! i"
        unfolding nth_map[OF i']
        using eqfE(1)[OF insts[unfolded len_ds, OF i[unfolded len_ts]]] 
        by (metis concat_sss i' len_ds len_sss map_nth nth_map)
    } 
    then show "\<forall>i<length ts. ts ! i = ?fhs ! i" by auto
  qed 
  note ts = this

  show "fill_holes c ts = fill_holes (fill_holes_mctxt c ds) (concat sss)"
    unfolding fill_holes_mctxt_fill_holes[OF len_ds nh] ts ..
qed      

fun compose_mctxt :: "('f, 'v) mctxt \<Rightarrow> nat \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt"
where
  "compose_mctxt C i Ci =
    fill_holes_mctxt C [(if i = j then Ci else MHole). j \<leftarrow> [0 ..< num_holes C]]"

lemma funas_mctxt_compose_mctxt [simp]:
  assumes "i < num_holes C"
  shows "funas_mctxt (compose_mctxt C i D) = funas_mctxt C \<union> funas_mctxt D"
proof -
  let ?Ds = "[(if i = j then D else MHole). j \<leftarrow> [0 ..< num_holes C]]"
  have "num_holes C = length ?Ds" by simp
  then show ?thesis using assms by (auto) (case_tac "i = x", simp_all)
qed

lemma compose_mctxt_sound:
  assumes s: "s =\<^sub>f (C, bef @ si # aft)"
  and si: "si =\<^sub>f (Ci, ts)"
  and i: "i = length bef"
  shows "s =\<^sub>f (compose_mctxt C i Ci, bef @ ts @ aft)"
proof -
  let ?Cs = "[ if i = j then Ci else MHole . j \<leftarrow> [0..<num_holes C]]"
  let ?ts = "bef @ si # aft"
  let ?sss = "[ [b]. b \<leftarrow> bef ] @ (ts # [ [a]. a \<leftarrow> aft])"

  have 
    l_Cs : "length ?Cs = num_holes C" and
    l_ts : "length ?ts = num_holes C" and
    l_sss : "length ?sss = num_holes C"
    unfolding length_append length_map list.size(4) using eqfE(2)[OF s] by auto

  have i_le_nh: "i < num_holes C" unfolding i eqfE(2)[OF s] length_append by (auto iff: trans_less_add1)
  have concat_sss: "concat ?sss = bef @ ts @ aft" by auto

  {
    fix j
    assume j: "j < i"
    then have j': "j < length [0..<num_holes C]" using i_le_nh length_upt by auto
    have "?sss!j = [bef!j]" by (metis append_Cons_nth_left i j length_map nth_map) 
    moreover have "?ts!j = bef!j" by (metis append_Cons_nth_left i j)
    moreover from nth_map[OF j'] j' j have "?Cs!j = MHole" by force
    ultimately have "?ts!j =\<^sub>f (?Cs!j,?sss!j)" using eqfI by auto
    } note j_le_i = this

  from i_le_nh have "?Cs!i = Ci" by auto
  moreover from i_le_nh have "?sss!i = ts" by (metis i length_map nth_append_length)
  moreover have "?ts!i = si" using nth_append_length i by auto
  ultimately have j_eq_i: "?ts!i =\<^sub>f (?Cs!i,?sss!i)" using si by auto

  {
    fix j
    assume j: "j > i" and j': "j < num_holes C"
    then have j'': "j < length [0..<num_holes C]" by auto
    
    have j''': "(j - i) - 1 < length aft" using 
      j'[unfolded eqfE(2)[OF s] length_append[of bef] list.size(4)] j 
      unfolding i by auto

    from nth_append[of "[ [b]. b \<leftarrow> bef ]" _ j, unfolded length_map[of _ bef] i[symmetric]] 
    have "?sss!j = (ts # [ [a]. a \<leftarrow> aft]) ! (j - i)"  using j by auto
    moreover have "... = [ [a]. a \<leftarrow> aft] ! ((j - i) - 1)" using nth_Cons_pos j by simp
    moreover have "... = [aft ! ((j - i) - 1)]" using j''' length_map nth_map by auto
    ultimately have sssj: "?sss!j = [aft ! ((j - i) - 1)]" by auto
    
    have Csj: "?Cs!j = MHole" using nth_map[OF j''] j'' j  by force

    have "?ts!j = (si # aft) ! (j - i)" unfolding nth_append[of bef] i[symmetric] using j by simp
    moreover have "... = aft ! ((j - i) - 1)" by (metis j neq0_conv nth_Cons' zero_less_diff)
    ultimately have "?ts!j = aft ! ((j - i) - 1)" by auto
    then have "?ts!j =\<^sub>f (?Cs!j,?sss!j)" using sssj Csj by auto
  } note j_gr_i = this  

  from j_le_i j_eq_i j_gr_i have "\<And> j. j < length ?Cs \<Longrightarrow> ?ts!j =\<^sub>f (?Cs!j,?sss!j)" 
    using l_Cs linorder_neqE_nat by metis
  
  from fill_holes_mctxt_sound[OF l_Cs l_sss l_ts this, unfolded concat_sss, folded compose_mctxt.simps]
  show ?thesis unfolding eqfE(1)[OF s] by simp
qed

fun mctxt_fill_partially_mctxts :: "('f, 'v) term list \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) mctxt list"
where 
  "mctxt_fill_partially_mctxts [] ts = map mctxt_of_term ts" |
  "mctxt_fill_partially_mctxts (s # ss) (t # ts) = 
    (if s = t then (MHole # mctxt_fill_partially_mctxts ss ts) 
    else (mctxt_of_term t # mctxt_fill_partially_mctxts (s # ss) ts))"

fun
  mctxt_fill_partially_fills ::
    "('f, 'v) term list \<Rightarrow> ('f, 'v) term list \<Rightarrow> ('f, 'v) term list list"
where 
  "mctxt_fill_partially_fills [] ts = map (const []) ts" |
  "mctxt_fill_partially_fills (s # ss) (t # ts) = 
    (if s = t then ([s] # mctxt_fill_partially_fills ss ts) 
    else ([] # mctxt_fill_partially_fills (s # ss) ts))"

lemma mctxt_fill_partially_mctxts_length [simp]:
  assumes "sublisteq ss ts"
  shows "length (mctxt_fill_partially_mctxts ss ts) = length ts"
  using assms by (induct rule: sublisteq_induct2, auto)

lemma mctxt_fill_partially_fills_length [simp]:
  assumes "sublisteq ss ts"
  shows "length (mctxt_fill_partially_fills ss ts) = length ts"
  using assms by (induct rule: sublisteq_induct2, auto)

lemma mctxt_fill_partially_numholes:
  assumes "sublisteq ss ts"
  shows "listsum [num_holes ci . ci \<leftarrow> mctxt_fill_partially_mctxts ss ts] = length ss"
proof (induct ss ts rule: sublisteq_induct2)
  case (goal3 s ss ts) 
  have ls_one: "\<And> as. listsum (1 # as) = listsum as + Suc 0" 
    by (metis One_nat_def Suc_eq_plus1 Suc_eq_plus1_left listsum_simps(2))
  from goal3 show ?case 
    unfolding mctxt_fill_partially_mctxts.simps list.size
    by (metis (full_types) One_nat_def Suc_eq_plus1 ls_one list.map(2) num_holes.simps(2))
next
  case (goal4 s ss t ts) 
  have ls_zero: "\<And> as. listsum (0 # as) = listsum as" by (metis listsum_simps(2) monoid_add_class.add.left_neutral) 
  have else: 
    "mctxt_fill_partially_mctxts (s # ss) (t # ts) = mctxt_of_term t # mctxt_fill_partially_mctxts (s # ss) ts" 
    using goal4(1) by auto
  show ?case 
    unfolding else list.map(2) num_holes_mctxt_of_term ls_zero goal4(5) ..
qed (auto iff: assms)

lemma mctxt_fill_partially_sound:
  assumes sl: "sublisteq ss ts"
  shows "\<And> i. i < length ts \<Longrightarrow> ts!i =\<^sub>f (mctxt_fill_partially_mctxts ss ts ! i, mctxt_fill_partially_fills ss ts ! i)"
proof (rule eqfI)
  let ?zipped= "zip (mctxt_fill_partially_mctxts ss ts) (mctxt_fill_partially_fills ss ts)"

  have l: "length ?zipped = length ts" 
    unfolding length_zip mctxt_fill_partially_mctxts_length[OF sl] mctxt_fill_partially_fills_length[OF sl] by auto

  have fh: "ts = map (\<lambda> (ci,tsi) . fill_holes ci tsi) ?zipped" 
  proof (induct ss ts rule: sublisteq_induct2)
    case (goal2 ts) then show ?case
    proof (induct ts, insert mctxt_of_term_fill_holes, auto) qed
  qed (insert sl, auto)

  have nh: "list_all (\<lambda> (ci,tsi) . num_holes ci = length tsi) ?zipped"
  proof (induct ss ts rule: sublisteq_induct2)
    case (goal2 ts) then show ?case
    proof (induct ts, insert num_holes_mctxt_of_term, auto) qed
  qed (insert sl, auto)

  { 
    fix i
    assume "i < length ts"
    then have 
      i1: "i < length (mctxt_fill_partially_mctxts ss ts)" and
      i2: "i < length (mctxt_fill_partially_fills ss ts)" 
      unfolding mctxt_fill_partially_mctxts_length[OF sl] mctxt_fill_partially_fills_length[OF sl] by auto
  } note i = this

  case goal1 
  then show ?case using fh nth_zip[OF i(1) i(2)]
    by (metis (lifting, no_types) goal1 list_update_id list_update_same_conv map_update split_conv)
  case goal2 then show ?case using nh[unfolded list_all_length] nth_zip[OF i(1) i(2)] 
    by (metis (lifting, full_types) l split_conv)
qed

lemma mctxt_fill_partially: 
  assumes ss: "sublisteq ss ts"
    and t: "t =\<^sub>f (c,ts)"
  shows "\<exists> d. t =\<^sub>f (d,ss)"
proof - 
  let ?ds = "mctxt_fill_partially_mctxts ss ts"
  let ?sss = "mctxt_fill_partially_fills ss ts"

  have "fill_holes c ts =\<^sub>f (fill_holes_mctxt c ?ds, concat ?sss)"
    using 
      fill_holes_mctxt_sound eqfE(2)[OF t,symmetric] mctxt_fill_partially_sound[OF ss] 
      mctxt_fill_partially_mctxts_length[OF ss] mctxt_fill_partially_fills_length[OF ss]
    by metis
  also have "concat ?sss = ss" proof (induct ss ts rule: sublisteq_induct2, insert ss, auto) qed
  ultimately show ?thesis by (metis eqfE(1) t)
qed

lemma fill_holes_mctxt_map_mctxt_of_term_conv [simp]:
  assumes "num_holes C = length ts"
  shows "fill_holes_mctxt C (map mctxt_of_term ts) = mctxt_of_term (fill_holes C ts)"
using assms
apply (induct C arbitrary: ts)
apply (auto simp: drop_map take_map)
apply (case_tac ts)
apply (auto)
by (metis length_map length_partition_by_nth nth_map nth_mem partition_holes_nth)

lemma fill_holes_mctxt_of_ctxt [simp]:
  "fill_holes (mctxt_of_ctxt C) [t] = C\<langle>t\<rangle>"
proof -
  have "C\<langle>t\<rangle> =\<^sub>f (mctxt_of_ctxt C, [t])" by (metis mctxt_of_ctxt)
  from eqfE [OF this] show ?thesis by simp
qed

definition
  "compose_cap_till P t i C =
    fill_holes_mctxt (cap_till P t) (map mctxt_of_term (take i (uncap_till P t)) @
      C # map mctxt_of_term (drop (Suc i) (uncap_till P t)))"

abbreviation "compose_cap_till_funas F \<equiv> compose_cap_till (if_Fun_in_set F)"

lemma fill_holes_compose_cap_till:
  assumes "i < num_holes (cap_till P s)" and "num_holes C = length ts"
  shows "fill_holes (compose_cap_till P s i C) ts =
    fill_holes (cap_till P s) (take i (uncap_till P s) @ fill_holes C ts # drop (Suc i) (uncap_till P s))"
  (is "_ = fill_holes _ ?ss")
proof -
  have "fill_holes (cap_till P s) ?ss =\<^sub>f
    (fill_holes_mctxt (cap_till P s) (map mctxt_of_term (take i (uncap_till P s)) @
      C # map mctxt_of_term (drop (Suc i) (uncap_till P s))),
      concat (map (\<lambda>_. []) (take i (uncap_till P s)) @ ts #
              map (\<lambda>_. []) (drop (Suc i) (uncap_till P s))))"
    (is "_ =\<^sub>f (fill_holes_mctxt _ ?ts, concat ?us)")
  proof (rule fill_holes_mctxt_sound)
    show "length ?ss = num_holes (cap_till P s)"
      using assms by simp
  next
    show "length ?ts = num_holes (cap_till P s)"
      using assms by simp
  next
    show "length ?us = num_holes (cap_till P s)"
      using assms by simp
  next
    fix j
    assume "j < length ?ts"
    with assms have j: "j < length (uncap_till P s)" by simp
    show "?ss ! j =\<^sub>f (?ts ! j, ?us ! j)"
      using assms and j by (cases "j = i") (auto simp: nth_append)
  qed
  note * = eqfE(1) [OF this]
  show ?thesis by (simp add: compose_cap_till_def *)
qed

lemma in_uncap_till_funas:
  assumes root: "root u = Some fn" "fn \<in> F"
    and "t = C\<langle>u\<rangle>"
  shows "\<exists>i < length (uncap_till_funas F t). \<exists>D. uncap_till_funas F t ! i = D\<langle>u\<rangle> \<and>
    mctxt_of_ctxt C = compose_cap_till_funas F t i (mctxt_of_ctxt D)"
using \<open>t = C\<langle>u\<rangle>\<close>
proof (induct t arbitrary: C)
  case (Var x)
  then show ?case using root by (cases C) (auto simp: wf_trs_def)
next
  case (Fun f ts)
  def [simp]: t \<equiv> "Fun f ts"
  show ?case
  proof (cases "(f, length ts) \<in> F")
    case True
    then show ?thesis using Fun.prems by (auto simp: compose_cap_till_def)
  next
    case False
    show ?thesis
    proof (cases C)
      case Hole
      then show ?thesis using Fun.prems and False and root by auto
    next
      case (More _ ss\<^sub>1 D _)
      moreover def j \<equiv> "length ss\<^sub>1"
      ultimately have j: "j < length ts" "ts ! j = D\<langle>u\<rangle>"
        and C: "C = More f (take j ts) D (drop (Suc j) ts)"
        using Fun.prems by (auto)
      then have "D\<langle>u\<rangle> \<in> set ts" by (auto simp: in_set_conv_nth)
      then obtain i and E
        where i: "i < length (uncap_till_funas F (D\<langle>u\<rangle>))" "uncap_till_funas F (D\<langle>u\<rangle>) ! i = E\<langle>u\<rangle>"
        and D: "mctxt_of_ctxt D = compose_cap_till_funas F (D\<langle>u\<rangle>) i (mctxt_of_ctxt E)"
        using Fun by blast
      obtain k where k: "take k (uncap_till_funas F t) =
        concat (map (uncap_till_funas F) (take j ts)) @ take i (uncap_till_funas F (D\<langle>u\<rangle>))"
        "k < length (uncap_till_funas F t)" "uncap_till_funas F t ! k = E\<langle>u\<rangle>"
        "drop (Suc k) ((uncap_till_funas F) t) = drop (Suc i) ((uncap_till_funas F) (D\<langle>u\<rangle>)) @ concat (map (uncap_till_funas F) (drop (Suc j) ts))"
        using False and i and j and take_nth_drop_concat [of j "map (uncap_till_funas F) ts" "(uncap_till_funas F) (D\<langle>u\<rangle>)" i "E\<langle>u\<rangle>"]
        by (auto simp: take_map drop_map)
      moreover have "mctxt_of_ctxt C = compose_cap_till_funas F t k (mctxt_of_ctxt E)"
      proof -
        have *: "compose_cap_till_funas F t k (mctxt_of_ctxt E) =
          fill_holes_mctxt (MFun f (map (cap_till_funas F) ts)) (concat (
            map (map mctxt_of_term \<circ> uncap_till_funas F) (take j ts) @
            (map mctxt_of_term (take i (uncap_till_funas F D\<langle>u\<rangle>)) @
            mctxt_of_ctxt E #
            map mctxt_of_term (drop (Suc i) (uncap_till_funas F D\<langle>u\<rangle>))) #
            map (map mctxt_of_term \<circ> uncap_till_funas F) (drop (Suc j) ts)))"
          (is "_ = fill_holes_mctxt _ (concat ?ss)")
          using False and k
          by (simp del: fill_holes_mctxt.simps add: compose_cap_till_def map_concat)
        also have "\<dots> = MFun f (map mctxt_of_term (take j ts) @ mctxt_of_ctxt D #
          map mctxt_of_term (drop (Suc j) ts))" (is "_ = MFun f ?ts")
        proof (rule fill_holes_mctxt_MFun)
          show "length (map (cap_till_funas F) ts) = length ?ts" using j by simp
        next
          show "length ?ss = length ?ts" by simp
        next
          fix n
          assume "n < length ?ts"
          then have n: "n < length ts" using j by simp
          show "num_holes (map (cap_till_funas F) ts ! n) = length (?ss ! n) \<and>
            fill_holes_mctxt (map (cap_till_funas F) ts ! n) (?ss ! n) = ?ts ! n"
          proof (cases "n = j")
            case False
            then have *: "?ss ! n = map mctxt_of_term (uncap_till_funas F (ts ! n))"
              "?ts ! n = mctxt_of_term (ts ! n)"
              using n and j by (auto simp: nth_append min_def)
            have "num_holes (map (cap_till_funas F) ts ! n) = length (?ss ! n)"
              using n by (simp add: * )
            moreover have "fill_holes_mctxt (map (cap_till_funas F) ts ! n) (?ss ! n) = ?ts ! n"
              using n by (auto simp: * )
            ultimately show ?thesis by blast
          next
            case True
            then have *: "?ss ! n =
              map mctxt_of_term (take i (uncap_till_funas F D\<langle>u\<rangle>)) @ mctxt_of_ctxt E #
                map mctxt_of_term (drop (Suc i) (uncap_till_funas F D\<langle>u\<rangle>))"
              "?ts ! n = mctxt_of_ctxt D"
              using n and j by (auto simp: nth_append)
            have "fill_holes_mctxt (map (cap_till_funas F) ts ! n) (?ss ! n) = ?ts ! n"
              unfolding * by (simp add: D compose_cap_till_def True j)
            moreover have "num_holes (map (cap_till_funas F) ts ! n) = length (?ss ! n)"
              unfolding * using i by (simp add: j True)
            ultimately show ?thesis by blast
          qed
        qed
        finally show ?thesis by (simp add: C)
      qed
      ultimately show ?thesis unfolding t_def by blast
    qed
  qed
qed

lemma uncap_till_funas_fill_holes_cancel [simp]:
  assumes "num_holes C = length ts" and "ground_mctxt C"
    and "funas_mctxt C \<subseteq> - F"
  shows "uncap_till_funas F (fill_holes C ts) = concat (map (uncap_till_funas F) ts)"
using assms
proof (induct C arbitrary: ts)
  case MHole
  then show ?case by (cases ts) simp_all
next
  case (MFun f Cs)
  let ?ts = "partition_holes ts Cs"
  let ?us = "partition_holes (map (uncap_till_funas F) ts) Cs"
  have *: "fill_holes (MFun f Cs) ts =
    Fun f (map (\<lambda>i. fill_holes (Cs ! i) (?ts ! i)) [0 ..< length Cs])"
    unfolding partition_holes_fill_holes_conv ..
  have "\<forall>i < length Cs. uncap_till_funas F (fill_holes (Cs ! i) (?ts ! i)) = concat (?us ! i)"
  proof (intro allI impI)
    fix i
    assume "i < length Cs"
    then have "Cs ! i \<in> set Cs" by simp
    from MFun.hyps [OF this, of "?ts ! i"] and MFun.prems and \<open>i < length Cs\<close>
      show "uncap_till_funas F (fill_holes (Cs ! i) (?ts ! i)) = concat (?us ! i)"
      by (auto iff: UN_subset_iff)
  qed
  then have **: "map (uncap_till_funas F \<circ> (\<lambda>i. fill_holes (Cs ! i) (?ts ! i))) [0..<length Cs] =
    map (concat \<circ> (\<lambda>i. (?us ! i))) [0 ..<length Cs]" by simp
  have ***: "listsum (map num_holes Cs) = length (map (uncap_till_funas F) ts)"
    using MFun.prems by simp
  show ?case
    using MFun.prems
    apply (simp add: * ** del: fill_holes.simps)
    apply (auto simp: o_def)
    unfolding map_upt_len_same_len_conv [OF length_partition_holes]
    by (simp add: concat_partition_by [OF ***])
qed simp

lemma uncap_till_funas_fill_holes_cap_till_funas [simp]:
  assumes "num_holes (cap_till_funas F s) = length ts"
  shows "uncap_till_funas F (fill_holes (cap_till_funas F s) ts) =
    concat (map (uncap_till_funas F) ts)"
  by (rule uncap_till_funas_fill_holes_cancel [OF assms ground_cap_till_funas, of F]) auto
   
lemma Ball_atLeast0LessThan_partition_holes_conv [simp]:
  "(\<forall>i \<in> {0 ..< length Cs}. \<forall>x \<in> set (partition_holes xs Cs ! i). P x) =
    (\<forall>x \<in> \<Union>(set (map set (partition_holes xs Cs))). P x)"
  using Ball_atLeast0LessThan_partition_by_conv [of "map num_holes Cs" xs] by simp

lemma ground_fill_holes_mctxt [simp]:
  "num_holes C = length Ds \<Longrightarrow>
    ground_mctxt (fill_holes_mctxt C Ds) \<longleftrightarrow> ground_mctxt C \<and> (\<forall>D \<in> set Ds. ground_mctxt D)"
proof (induct C arbitrary: Ds)
  case MHole
  then show ?case by (cases Ds) simp_all
next
  case (MFun f Cs)
  then have *: "(\<forall>i\<in>{0..<length Cs}.
    ground_mctxt (fill_holes_mctxt (Cs ! i) (partition_holes Ds Cs ! i))) =
    (\<forall>i\<in>{0..<length Cs}.
      ground_mctxt (Cs ! i) \<and> (\<forall>a\<in>set (partition_holes Ds Cs ! i). ground_mctxt a))"
    and **: "listsum (map num_holes Cs) = length Ds"
    by simp_all
  show ?case
    unfolding partition_holes_fill_holes_mctxt_conv
    by (simp add: * ball_conj_distrib Ball_set_partition_by [OF **])
qed simp

lemma concat_map_uncap_till_funas_map_subst_apply_uncap_till_funas [simp]:
  "concat (map (uncap_till_funas F) (map (\<lambda>s. s \<cdot> \<sigma>) (uncap_till_funas F t))) = uncap_till_funas F (t \<cdot> \<sigma>)"
proof (induct t)
  case (Fun f ts)
  then have *: "map (uncap_till_funas F \<circ> (\<lambda>t. t \<cdot> \<sigma>)) ts =
    map concat (map (map (uncap_till_funas F) \<circ> map (\<lambda>s. s \<cdot> \<sigma>) \<circ> uncap_till_funas F) ts)" by simp
  show ?case
    by (simp add: * map_concat concat_map_concat [symmetric])
qed simp

lemma concat_uncap_till_subst_conv:
  "concat (map (\<lambda>i. uncap_till_funas F ((uncap_till_funas F t ! i) \<cdot> \<sigma>)) [0 ..< length (uncap_till_funas F t)]) =
    uncap_till_funas F (t \<cdot> \<sigma>)"
proof -
  have "concat (map (uncap_till_funas F) (map (\<lambda>i.
    (uncap_till_funas F t ! i) \<cdot> \<sigma>) [0 ..< length (uncap_till_funas F t)])) = uncap_till_funas F (t \<cdot> \<sigma>)"
    unfolding map_upt_len_conv [of "\<lambda>s. s \<cdot> \<sigma>" "uncap_till_funas F t"]
    unfolding concat_map_uncap_till_funas_map_subst_apply_uncap_till_funas ..
  then show ?thesis by (simp add: o_def)
qed

lemma the_root_uncap_till_funas:
  "is_Fun t \<Longrightarrow> the (root t) \<in> F \<Longrightarrow> uncap_till_funas F t = [t]"
  by (cases t) simp_all

lemma funas_cap_till_subset:
  "funas_mctxt (cap_till P t) \<subseteq> funas_term t"
  by (induct t) auto

lemma funas_uncap_till_subset:
  "s \<in> set (uncap_till P t) \<Longrightarrow> funas_term s \<subseteq> funas_term t"
proof (induct t arbitrary: s)
  case (Fun f ts)
  then show ?case by (cases "P (Fun f ts)") auto
qed simp

lemma ground_mctxt_subst_apply_context [simp]:
  "ground_mctxt C \<Longrightarrow> C \<cdot>mc \<sigma> = C"
  by (induct C) (simp_all add: map_idI)

lemma vars_term_fill_holes [simp]:
  "num_holes C = length ts \<Longrightarrow> ground_mctxt C \<Longrightarrow>
    vars_term (fill_holes C ts) = \<Union>(vars_term ` set ts)"
proof (induct C arbitrary: ts)
  case MHole
  then show ?case by (cases ts) simp_all
next
  case (MFun f Cs)
  then have *: "length (partition_holes ts Cs) = length Cs" by simp
  let ?f = "\<lambda>x. \<Union>y \<in> set x. vars_term y"
  show ?case
    using MFun
    unfolding partition_holes_fill_holes_conv
    by (simp add: UN_upt_len_conv [OF *, of ?f] UN_set_partition_by)
qed simp

subsection \<open>@{text "show"} Instance for Multihole Contexts\<close>

text \<open>
  Show-function for multihole contexts that is parametrized over show-functions for
  function symbols and variables.
\<close>
fun shows_mctxt' :: "('f \<Rightarrow> shows) \<Rightarrow> ('v \<Rightarrow> shows) \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> shows"
where
  "shows_mctxt' fun var MHole = shows_string ''[]''" |
  "shows_mctxt' fun var (MVar x) = var x" |
  "shows_mctxt' fun var (MFun f Cs) =
    (fun f +@+ shows_list_gen id [] ''('' '', '' '')'' (map (shows_mctxt' fun var) Cs))"

instantiation mctxt :: ("show", "show") "show"
begin

abbreviation "shows_mctxt C \<equiv> shows_mctxt' shows shows C"
definition "shows_prec (p::nat) C = shows_mctxt C"

lemma shows_prec_mctxt_append [show_law_simps]:
  "shows_prec p (C :: ('a, 'b) mctxt) (r @ s) = shows_prec p C r @ s"
  by (induct C arbitrary: r s) (simp_all add: shows_prec_mctxt_def show_law_simps)

definition "shows_list (Cs :: ('a, 'b) mctxt list) = showsp_list shows_prec 0 Cs"

instance by (default) (simp_all add: shows_list_mctxt_def show_law_simps)

end


subsection \<open>Semilattice Structures\<close>

instantiation mctxt :: (type, type) inf
begin

fun inf_mctxt :: "('a, 'b) mctxt \<Rightarrow> ('a, 'b) mctxt \<Rightarrow> ('a, 'b) mctxt"
where
  "MHole \<sqinter> D = MHole" |
  "C \<sqinter> MHole = MHole" |
  "MVar x \<sqinter> MVar y = (if x = y then MVar x else MHole)" |
  "MFun f Cs \<sqinter> MFun g Ds =
    (if f = g \<and> length Cs = length Ds then MFun f (map (split op \<sqinter>) (zip Cs Ds))
    else MHole)" |
  "C \<sqinter> D = MHole"

instance ..

end

lemma inf_mctxt_idem [simp]:
  fixes C :: "('f, 'v) mctxt"
  shows "C \<sqinter> C = C"
  by (induct C) (auto simp: zip_same_conv_map intro: map_idI)

lemma inf_mctxt_MHole2 [simp]:
  "C \<sqinter> MHole = MHole"
  by (induct C) simp_all

lemma inf_mctxt_comm [ac_simps]:
  "(C :: ('f, 'v) mctxt) \<sqinter> D = D \<sqinter> C"
  by (induct C D rule: inf_mctxt.induct) (fastforce simp: in_set_conv_nth intro!: nth_equalityI)+

lemma inf_mctxt_assoc [ac_simps]:
  fixes C :: "('f, 'v) mctxt"
  shows "C \<sqinter> D \<sqinter> E = C \<sqinter> (D \<sqinter> E)"
  apply (induct C D arbitrary: E rule: inf_mctxt.induct)
  apply (auto simp: )
  apply (case_tac E, auto)+
  apply (fastforce simp: in_set_conv_nth intro!: nth_equalityI)
  apply (case_tac E, auto)+
done

instantiation mctxt :: (type, type) order
begin

definition "(C :: ('a, 'b) mctxt) \<le> D \<longleftrightarrow> C \<sqinter> D = C"
definition "(C :: ('a, 'b) mctxt) < D \<longleftrightarrow> C \<le> D \<and> \<not> D \<le> C"

instance
  by (default, simp_all add: less_eq_mctxt_def less_mctxt_def ac_simps, metis inf_mctxt_assoc)

end

instance mctxt :: (type, type) semilattice_inf
  apply (default)
  apply (auto simp: less_eq_mctxt_def inf_mctxt_assoc [symmetric])
  apply (metis inf_mctxt_comm inf_mctxt_assoc inf_mctxt_idem)+
done

fun inf_mctxt_args :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt list"
where
  "inf_mctxt_args MHole D = [MHole]" |
  "inf_mctxt_args C MHole = [C]" |
  "inf_mctxt_args (MVar x) (MVar y) = (if x = y then [] else [MVar x])" |
  "inf_mctxt_args (MFun f Cs) (MFun g Ds) =
    (if f = g \<and> length Cs = length Ds then concat (map (split inf_mctxt_args) (zip Cs Ds))
    else [MFun f Cs])" |
  "inf_mctxt_args C D = [C]"

lemma inf_mctxt_args_MHole2 [simp]:
  "inf_mctxt_args C MHole = [C]"
  by (cases C) simp_all

lemma fill_holes_mctxt_replicate_MHole [simp]:
  "fill_holes_mctxt C (replicate (num_holes C) MHole) = C"
using assms
proof (induct C)
  case (MFun f Cs)
  then have "\<forall>i < length Cs.
    fill_holes_mctxt (Cs ! i) (replicate (num_holes (Cs ! i)) MHole) = Cs ! i"
    by (force simp: set_zip)
  then show ?case
    by (auto intro!: nth_equalityI)
       (metis drop_replicate length_partition_holes_nth length_replicate partition_holes_nth take_replicate)
qed auto

lemma num_holes_inf_mctxt:
  "num_holes (C \<sqinter> D) = length (inf_mctxt_args C D)"
  by (induct C D rule: inf_mctxt.induct)
     (auto simp: in_set_zip length_concat intro!: arg_cong [of _ _ listsum])

lemma length_inf_mctxt_args:
  "length (inf_mctxt_args D C) = length (inf_mctxt_args C D)"
  by (metis inf.commute num_holes_inf_mctxt)

(*TODO: move*)
lemma less_length_concat:
  assumes "i < length (concat xs)"
  shows "\<exists>m n.
    i = listsum (map length (take m xs)) + n \<and>
    m < length xs \<and> n < length (xs ! m) \<and> concat xs ! i = xs ! m ! n"
using assms
proof (induct xs arbitrary: i rule: length_induct)
  case (1 xs)
  then show ?case
  proof (cases xs)
    case (Cons y ys)
    note [simp] = this
    { assume *: "i < length y"
      with 1 obtain n where "i = n" and "n < length y"
        and "y ! i = y ! n" by simp
      then have "i = listsum (map length (take 0 xs)) + n"
        and "0 < length xs" and "n < length (xs ! 0)"
        and "concat xs ! i = xs ! 0 ! n"
        using * by (auto simp: nth_append)
      then have ?thesis by blast }
    moreover
    { assume *: "i \<ge> length y"
      def j \<equiv> "i - length y"
      then have "length ys < length xs" "j < length (concat ys)"
        using * and "1.prems" by auto
      with 1 obtain m n where "j = listsum (map length (take m ys)) + n"
        and "m < length ys" and "n < length (ys ! m)"
        and "concat ys ! j = ys ! m ! n" by blast
      then have "i = listsum (map length (take (Suc m) xs)) + n"
        and "Suc m < length xs" and "n < length (xs ! Suc m)"
        and "concat xs ! i = xs ! Suc m ! n"
        using * by (simp_all add: j_def nth_append)
      then have ?thesis by blast }
    ultimately show ?thesis by force
  qed simp
qed

lemma inf_mctxt_args_same [simp]:
  "inf_mctxt_args C C = replicate (num_holes C) MHole"
proof (induct C)
  case (MFun f Cs)
  have *: "\<And>C. num_holes C = length (inf_mctxt_args C C)"
    using num_holes_inf_mctxt [of C C for C] by auto
  let ?xs = "map (split inf_mctxt_args) (zip Cs Cs)"
  have "\<forall>i < length Cs.
    inf_mctxt_args (Cs ! i) (Cs ! i) = replicate (num_holes (Cs ! i)) MHole" using MFun by auto
  then have "\<forall>i < length ?xs. \<forall>j < length (?xs ! i). ?xs ! i ! j = MHole" by auto
  then have "\<forall>i < length (concat ?xs). concat ?xs ! i = MHole" by (metis nth_concat_two_lists) 
  then show ?case by (auto simp: * intro!: nth_equalityI)
qed simp_all

lemma inf_mctxt_inf_mctxt_args:
  "fill_holes_mctxt (C \<sqinter> D) (inf_mctxt_args C D) = C"
proof (induct C D rule: inf_mctxt.induct)
  case (4 f Cs g Ds)
  then show ?case
  proof (cases "f = g \<and> length Cs = length Ds")
    case True
    with 4 have "\<forall>i < length Cs.
      fill_holes_mctxt (Cs ! i \<sqinter> Ds ! i) (inf_mctxt_args (Cs ! i) (Ds ! i)) = Cs ! i"
      by (force simp: set_zip)
    moreover have "partition_holes (concat (map (split inf_mctxt_args) (zip Cs Ds)))
      (map (split op \<sqinter>) (zip Cs Ds)) = map (split inf_mctxt_args) (zip Cs Ds)"
      by (rule partition_by_concat_id) (simp_all add: num_holes_inf_mctxt)
    ultimately show ?thesis
      using fill_holes_mctxt.simps [simp del]
      by (auto simp: partition_holes_fill_holes_mctxt_conv intro!: nth_equalityI)
  qed auto
qed auto

lemma inf_mctxt_inf_mctxt_args2:
  "fill_holes_mctxt (C \<sqinter> D) (inf_mctxt_args D C) = D"
  unfolding inf_mctxt_comm [of C D] by (rule inf_mctxt_inf_mctxt_args)

instantiation mctxt :: (type, type) sup
begin

fun sup_mctxt :: "('a, 'b) mctxt \<Rightarrow> ('a, 'b) mctxt \<Rightarrow> ('a, 'b) mctxt"
where
  "MHole \<squnion> D = D" |
  "C \<squnion> MHole = C" |
  "MVar x \<squnion> MVar y = (if x = y then MVar x else undefined)" |
  "MFun f Cs \<squnion> MFun g Ds =
    (if f = g \<and> length Cs = length Ds then MFun f (map (split op \<squnion>) (zip Cs Ds))
    else undefined)" |
  "(C :: ('a, 'b) mctxt) \<squnion> D = undefined"

instance ..

end

lemma sup_mctxt_idem [simp]:
  fixes C :: "('f, 'v) mctxt"
  shows "C \<squnion> C = C"
  by (induct C) (auto simp: zip_same_conv_map intro: map_idI)

lemma sup_mctxt_MHole [simp]: "C \<squnion> MHole = C"
  by (induct C) simp_all

lemma sup_mctxt_comm [ac_simps]:
  fixes C :: "('f, 'v) mctxt"
  shows "C \<squnion> D = D \<squnion> C"
  by (induct C D rule: sup_mctxt.induct) (fastforce simp: in_set_conv_nth intro!: nth_equalityI)+

text \<open>
  @{const sup} is defined on compatible multihole-contexts.
  Note that compatibility is not transitive.
\<close>
inductive_set comp_mctxt :: "(('a, 'b) mctxt \<times> ('a, 'b) mctxt) set"
where
  MHole1: "(MHole, D) \<in> comp_mctxt" |
  MHole2: "(C, MHole) \<in> comp_mctxt" |
  MVar: "x = y \<Longrightarrow> (MVar x, MVar y) \<in> comp_mctxt" |
  MFun: "f = g \<Longrightarrow> length Cs = length Ds \<Longrightarrow> \<forall>i < length Ds. (Cs ! i, Ds ! i) \<in> comp_mctxt \<Longrightarrow>
    (MFun f Cs, MFun g Ds) \<in> comp_mctxt"

lemma comp_mctxt_refl:
  "(C, C) \<in> comp_mctxt"
  by (induct C) (auto intro: comp_mctxt.intros)

lemma comp_mctxt_sym:
  assumes "(C, D) \<in> comp_mctxt"
  shows "(D, C) \<in> comp_mctxt"
  using assms by (induct) (auto intro: comp_mctxt.intros)

lemma sup_mctxt_assoc [ac_simps]:
  assumes "(C, D) \<in> comp_mctxt" and "(D, E) \<in> comp_mctxt"
  shows "C \<squnion> D \<squnion> E = C \<squnion> (D \<squnion> E)"
  using assms by (induct C D arbitrary: E) (auto elim!: comp_mctxt.cases intro!: nth_equalityI)

text \<open>
  No instantiation to @{class semilattice_sup} possible, since @{const sup} is only
  partially defined on terms (e.g., it is not associative in general).
\<close>

interpretation mctxt_order_bot: order_bot MHole "op \<le>" "op <"
  by (default) (simp add: less_eq_mctxt_def)

lemma sup_mctxt_ge1 [simp]:
  assumes "(C, D) \<in> comp_mctxt"
  shows "C \<le> C \<squnion> D"
  using assms by (induct C D) (auto simp: less_eq_mctxt_def intro: nth_equalityI)

lemma sup_mctxt_ge2 [simp]:
  assumes "(C, D) \<in> comp_mctxt"
  shows "D \<le> C \<squnion> D"
  using assms by (induct) (auto simp: less_eq_mctxt_def intro: nth_equalityI)

lemma sup_mctxt_least:
  assumes "(D, E) \<in> comp_mctxt"
    and "D \<le> C" and "E \<le> C"
  shows "D \<squnion> E \<le> C"
using assms
apply (induct arbitrary: C)
apply (auto simp: less_eq_mctxt_def elim!: inf_mctxt.elims intro!: nth_equalityI)
apply (metis (erased, lifting) length_map nth_map nth_zip split_conv)
apply (case_tac "fb = gb \<and> length Csb = length Dsb", simp_all)+
done

lemma inf_mctxt_args_MHole:
  assumes "(C, D) \<in> comp_mctxt" and "i < length (inf_mctxt_args C D)"
  shows "inf_mctxt_args C D ! i = MHole \<or> inf_mctxt_args D C ! i = MHole"
using assms
proof (induct C D arbitrary: i)
  case (MHole2 C)
  then show ?case by (cases C) simp_all
next
  case (MFun f g Cs Ds)
  then have [simp]: "f = g" "length Ds = length Cs" by auto
  let ?xs = "map (split inf_mctxt_args) (zip Cs Ds)"
  let ?ys = "map (split inf_mctxt_args) (zip Ds Cs)"
  obtain m and n where *: "i = listsum (map length (take m ?xs)) + n"
    "m < length Cs" "n < length (inf_mctxt_args (Cs ! m) (Ds ! m))"
    and "inf_mctxt_args (MFun f Cs) (MFun g Ds) ! i = inf_mctxt_args (Cs ! m) (Ds ! m) ! n"
    using MFun.prems by (auto dest: less_length_concat)
  moreover have "concat ?ys ! i = (map (split inf_mctxt_args) (zip Ds Cs)) ! m ! n"
      by (rule concat_nth)
         (auto intro: arg_cong [of _ _ listsum] simp: * map_nth_eq_conv length_inf_mctxt_args)
  ultimately show ?case using MFun(3) by simp
qed auto

lemma fill_holes_mctxt_Nil [simp]:
  "fill_holes_mctxt C [] = C"
  by (induct C) (auto intro: nth_equalityI)

(*TODO: move to theory about parallel rewriting? (Which is currently part of Trs)*)
text \<open>Parallel rewriting is closed under multihole-contexts.\<close>
lemma par_rstep_mctxt:
  assumes "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> par_rstep R"
  shows "(s, t) \<in> par_rstep R"
proof -
  have [simp]: "length ss = length ts" using assms by (auto dest!: eqfE)
  have [simp]: "t = fill_holes C ts" using assms by (auto dest: eqfE)
  have "(s, fill_holes C ts) \<in> par_rstep R"
    using assms by (intro eqf_all_ctxt_closed_step [of UNIV _ s C ss, THEN conjunct1]) auto
  then show ?thesis by simp
qed

(*TODO: move to theory about parallel rewriting?*)
lemma par_rstep_mctxtD:
  assumes "(s, t) \<in> par_rstep R"
  shows "\<exists>C ss ts. s =\<^sub>f (C, ss) \<and> t =\<^sub>f (C, ts) \<and> (\<forall>i<length ss. (ss ! i, ts ! i) \<in> rrstep R)"
  (is "\<exists>C ss ts. ?P s t C ss ts")
using assms
proof (induct)
  case (root_step s t \<sigma>)
  then have "(s \<cdot> \<sigma>, t \<cdot> \<sigma>) \<in> rrstep R" by auto
  moreover have "s \<cdot> \<sigma> =\<^sub>f (MHole, [s \<cdot> \<sigma>])" and "t \<cdot> \<sigma> =\<^sub>f (MHole, [t \<cdot> \<sigma>])" by auto
  ultimately show ?case by force
next
  case (par_step_var x)
  have "Var x =\<^sub>f (MVar x, [])" by auto
  then show ?case by force
next
  case (par_step_fun ts ss f)
  then have "\<forall>i<length ts. \<exists>x. ?P (ss ! i) (ts ! i) (fst x) (fst (snd x)) (snd (snd x))" by force
  then obtain g where "\<forall>i<length ts. ?P (ss ! i) (ts ! i) (fst (g i)) (fst (snd (g i))) (snd (snd (g i)))"
    unfolding choice_iff' by blast
  moreover
  def Cs \<equiv> "map (\<lambda>i. fst (g i)) [0 ..< length ts]"
    and us \<equiv> "map (\<lambda>i. fst (snd (g i))) [0 ..< length ts]"
    and vs \<equiv> "map (\<lambda>i. snd (snd (g i))) [0 ..< length ts]"
  ultimately have [simp]: "length Cs = length ts"
    "length us = length ts" "length vs = length ts"
    and *: "\<forall>i<length us. ss ! i =\<^sub>f (Cs ! i, us ! i) \<and> ts ! i =\<^sub>f (Cs ! i, vs ! i) \<and>
      (\<forall>j<length (us ! i). (us ! i ! j, vs ! i ! j) \<in> rrstep R)"
    by simp_all
  def C \<equiv> "MFun f Cs"
  have "Fun f ss =\<^sub>f (C, concat us)" and "Fun f ts =\<^sub>f (C, concat vs)"
    using * by (auto simp: C_def \<open>length ss = length ts\<close> intro: eqf_MFunI)
  moreover have "\<forall>i<length (concat us). (concat us ! i, concat vs ! i) \<in> rrstep R"
    using * by (intro concat_all_nth) (auto dest!: eqfE)
  ultimately show ?case by blast
qed

lemma rsteps_mctxt:
  assumes "s =\<^sub>f (C, ss)" and "t =\<^sub>f (C, ts)"
    and "\<forall>i<length ss. (ss ! i, ts ! i) \<in> (rstep R)\<^sup>*"
  shows "(s, t) \<in> (rstep R)\<^sup>*"
proof -
  have [simp]: "length ss = length ts" using assms by (auto dest!: eqfE)
  have [simp]: "t = fill_holes C ts" using assms by (auto dest: eqfE)
  have "(s, fill_holes C ts) \<in> (rstep R)\<^sup>*"
    using assms by (intro eqf_all_ctxt_closed_step [of UNIV _ s C ss, THEN conjunct1]) auto
  then show ?thesis by simp
qed

fun sup_mctxt_args :: "('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt \<Rightarrow> ('f, 'v) mctxt list"
where
  "sup_mctxt_args MHole D = [D]" |
  "sup_mctxt_args C MHole = replicate (num_holes C) MHole" |
  "sup_mctxt_args (MVar x) (MVar y) = (if x = y then [] else undefined)" |
  "sup_mctxt_args (MFun f Cs) (MFun g Ds) =
    (if f = g \<and> length Cs = length Ds then concat (map (split sup_mctxt_args) (zip Cs Ds))
    else undefined)" |
  "sup_mctxt_args C D = undefined"

lemma sup_mctxt_args_MHole2 [simp]:
  "sup_mctxt_args C MHole = replicate (num_holes C) MHole"
  by (cases C) simp_all

lemma num_holes_sup_mctxt_args:
  assumes "(C, D) \<in> comp_mctxt"
  shows "num_holes C = length (sup_mctxt_args C D)"
  using assms by (induct) (auto simp: length_concat intro!: arg_cong [of _ _ listsum] nth_equalityI)

lemma sup_mctxt_sup_mctxt_args:
  assumes "(C, D) \<in> comp_mctxt"
  shows "fill_holes_mctxt C (sup_mctxt_args C D) = C \<squnion> D"
using assms
proof (induct)
  note fill_holes_mctxt.simps [simp del]
  case (MFun f g Cs Ds)
  then show ?case
  proof (cases "f = g \<and> length Cs = length Ds")
    case True
    with MFun have "\<forall>i < length Cs.
      fill_holes_mctxt (Cs ! i) (sup_mctxt_args (Cs ! i) (Ds ! i)) = Cs ! i \<squnion> Ds ! i"
      and *: "\<forall>i < length Cs. (Cs ! i, Ds ! i) \<in> comp_mctxt" by (force simp: set_zip)+
    moreover have "partition_holes (concat (map (split sup_mctxt_args) (zip Cs Ds)))
      Cs = map (split sup_mctxt_args) (zip Cs Ds)"
      using True and * by (intro partition_by_concat_id) (auto simp: num_holes_sup_mctxt_args)
    ultimately show ?thesis
      using * and True by (auto simp: partition_holes_fill_holes_mctxt_conv intro!: nth_equalityI)
  qed auto
qed auto

lemma sup_mctxt_args:
  assumes "(C, D) \<in> comp_mctxt"
  shows "sup_mctxt_args C D = inf_mctxt_args (C \<squnion> D) C"
  using assms by (induct) (auto intro!: arg_cong [of _ _ concat] nth_equalityI)

lemma term_for_mctxt:
  fixes C :: "('f, 'v) mctxt"
  obtains t and ts where "t =\<^sub>f (C, ts)"
proof -
  obtain ts :: "('f, 'v) term list" where "num_holes C = length ts" by (metis Ex_list_of_length)
  then have "fill_holes C ts =\<^sub>f (C, ts)" by blast
  show ?thesis by (default) fact
qed

lemma comp_mctxt_eqfE:
  assumes "(C, D) \<in> comp_mctxt"
  obtains s and ss and ts where "s =\<^sub>f (C, ss)" and "s =\<^sub>f (D, ts)"
proof -
  obtain u and us where "u =\<^sub>f (C \<squnion> D, us)" by (metis term_for_mctxt)
  then have u: "u = fill_holes (C \<squnion> D) us"
    and *: "length us = num_holes (C \<squnion> D)" by (auto dest: eqfE)
  def Cs \<equiv> "sup_mctxt_args C D"
  and Ds \<equiv> "sup_mctxt_args D C"
  then have sup1: "C \<squnion> D = fill_holes_mctxt C Cs" and sup2: "C \<squnion> D = fill_holes_mctxt D Ds"
    using assms by (auto simp: sup_mctxt_sup_mctxt_args comp_mctxt_sym ac_simps)
  then have u1: "u = fill_holes (fill_holes_mctxt C Cs) us"
    and u2: "u = fill_holes (fill_holes_mctxt D Ds) us" by (simp_all add: u)
  def ss \<equiv> "map (\<lambda>i. fill_holes (Cs ! i) (partition_holes us Cs ! i)) [0 ..< num_holes C]"
  and ts \<equiv> "map (\<lambda>i. fill_holes (Ds ! i) (partition_holes us Ds ! i)) [0 ..< num_holes D]"
  have "u = fill_holes C ss"
    using assms
     by (simp add: * u1 sup1 ss_def fill_holes_mctxt_fill_holes Cs_def num_holes_sup_mctxt_args)
  moreover have "u = fill_holes D ts"
    using assms [THEN comp_mctxt_sym]
    by (simp add: * u2 sup2 ts_def fill_holes_mctxt_fill_holes Ds_def num_holes_sup_mctxt_args)
  ultimately have "u =\<^sub>f (C, ss)" and "u =\<^sub>f (D, ts)" by (auto simp: ss_def ts_def)
  show ?thesis by (rule) fact+
qed

lemma eqf_comp_mctxt:
  assumes "s =\<^sub>f (C, ss)" and "s =\<^sub>f (D, ts)"
  shows "(C, D) \<in> comp_mctxt"
using assms
proof (induct s arbitrary: C D ss ts)
  case (Var x C D)
  then show ?case
    by (cases C D rule: mctxt.exhaust [case_product mctxt.exhaust])
       (auto simp: eq_fill.simps intro: comp_mctxt.intros)
next
  case (Fun f ss C D us vs)
  { fix Cs and Ds
    assume *: "C = MFun f Cs" "D = MFun f Ds" and **: "length Cs = length Ds"
    have ?case
    proof (unfold *, intro comp_mctxt.MFun [OF refl **] allI impI)
      fix i
      assume "i < length Ds"
      then show "(Cs ! i, Ds ! i) \<in> comp_mctxt"
        using Fun by (auto simp: * ** elim!: eqf_MFunE) (metis nth_mem)
    qed }
  with Fun.prems show ?case
    by (cases C D rule: mctxt.exhaust [case_product mctxt.exhaust])
       (auto simp: eq_fill.simps dest: map_eq_imp_length_eq intro: comp_mctxt.intros)
qed

lemma comp_mctxt_iff:
  "(C, D) \<in> comp_mctxt \<longleftrightarrow> (\<exists>s ss ts. s =\<^sub>f (C, ss) \<and> s =\<^sub>f (D, ts))"
  by (blast elim!: comp_mctxt_eqfE intro: eqf_comp_mctxt)

lemma hole_poss_parallel_pos [simp]:
  assumes "p \<in> hole_poss C" and "q \<in> hole_poss C" and "p \<noteq> q"
  shows "parallel_pos p q"
using assms by (induct C arbitrary: p q) (fastforce dest!: nth_mem)+

lemma eq_fill_induct [consumes 1, case_names MHole MVar MFun]:
  assumes "t =\<^sub>f (C, ts)"
    and "\<And>t. P t MHole [t]"
    and "\<And>x. P (Var x) (MVar x) []"
    and "\<And>f ss Cs ts. \<lbrakk>length Cs = length ss; listsum (map num_holes Cs) = length ts;
      \<forall>i < length ss. ss ! i =\<^sub>f (Cs ! i, partition_holes ts Cs ! i) \<and>
        P (ss ! i) (Cs ! i) (partition_holes ts Cs ! i)\<rbrakk>
      \<Longrightarrow> P (Fun f ss) (MFun f Cs) ts"
  shows "P t C ts"
using assms(1)
proof (induct t arbitrary: C ts)
  case (Var x)
  then show ?case
    using assms(2, 3) by (cases C; cases ts) (auto elim: eq_fill.cases)
next
  case (Fun f ss C ts)
  { assume "C = MHole" and "ts = [Fun f ss]"
    with Fun.hyps have ?case using assms(2) by auto }
  moreover
  { fix Cs
    assume C: "C = MFun f Cs" and "listsum (map num_holes Cs) = length ts"
      and "length Cs = length ss"
      and "Fun f ss = fill_holes (MFun f Cs) ts"
    moreover then have "\<forall>i < length ss. ss ! i =\<^sub>f (Cs ! i, partition_holes ts Cs ! i)"
      apply (auto simp del: fill_holes.simps
                  simp: partition_holes_fill_holes_conv intro!: eq_fill.intros)
      by (metis (no_types, lifting) add.left_neutral length_map length_upt nth_map_upt)
    moreover with Fun.hyps(1) have "\<forall>i < length ss. 
      P (ss ! i) (Cs ! i) (partition_holes ts Cs ! i)" by auto
    ultimately have ?case using assms(4) [of Cs ss ts f] by auto }
  ultimately show ?case
    using Fun.prems by (elim eq_fill.cases) (auto, cases C; cases ts, auto)
qed

lemma hole_poss_subset_poss:
  assumes "s =\<^sub>f (C, ss)"
  shows "hole_poss C \<subseteq> poss s"
using assms by (induct rule: eq_fill_induct) auto

fun hole_num
where
  "hole_num \<epsilon> MHole = 0" |
  "hole_num (i <# q) (MFun f Cs) = listsum (map num_holes (take i Cs)) + hole_num q (Cs ! i)"

(*TODO: move*)
lemma concat_nth_length:
  "i < length uss \<Longrightarrow> j < length (uss ! i) \<Longrightarrow>
    listsum (map length (take i uss)) + j < length (concat uss)"
by (induct uss arbitrary: i j) (simp, case_tac i, auto)

lemma hole_poss_nth_subt_at:
  assumes "t =\<^sub>f (C, ts)" and "p \<in> hole_poss C"
  shows "hole_num p C < length ts \<and> t |_ p = ts ! hole_num p C"
using assms
proof (induct arbitrary: p rule: eq_fill_induct)
  case (MFun f ss Cs ts)
  let ?ts = "partition_holes ts Cs"
  from MFun obtain i and q where [simp]: "p = i <# q"
    and "i < length ss" and "q \<in> hole_poss (Cs ! i)" by auto
  with MFun.hyps have "ss ! i =\<^sub>f (Cs ! i, ?ts ! i)"
    and j: "hole_num q (Cs ! i) < length (?ts ! i)" (is "?j < length _")
    and *: "?ts ! i ! hole_num q (Cs ! i) = ss ! i |_ q"
    by auto
  let ?k = "listsum (map length (take i ?ts)) + ?j"
  have "i < length ?ts" using \<open>i < length ss\<close> and MFun by auto
  with partition_by_nth_nth [OF this j] and MFun and concat_nth_length [OF this j]
    have "?ts ! i ! ?j = ts ! ?k" and "?k < length ts" by (auto)
  moreover with * have "ts ! ?k = Fun f ss |_ p" using \<open>i < length ss\<close> by simp
  ultimately show ?case using MFun.hyps(2) by (auto simp: take_map [symmetric])
qed auto

lemma eqf_Fun_MFun:
  assumes "Fun f ss =\<^sub>f (MFun g Cs, ts)"
  shows "g = f \<and> length Cs = length ss \<and> listsum (map num_holes Cs) = length ts \<and>
    (\<forall>i < length ss. ss ! i =\<^sub>f (Cs ! i, partition_holes ts Cs ! i))"
using assms by (induct "Fun f ss" "MFun g Cs" ts rule: eq_fill_induct) auto

lemma fill_holes_eq_Var_cases:
  assumes "num_holes C = length ts"
    and "fill_holes C ts = Var x"
  obtains "C = MHole \<and> ts = [Var x]" | "C = MVar x \<and> ts = []"
using assms by (induct C; cases ts) auto

end
